"""
Module for writing deepmd training sets from ab-initio calculation results.
"""
import os
import glob
import json
import numpy as np
from ase.db import connect
from ase.io import read
from ase.data import atomic_numbers
from ase.io.formats import string2index
from sklearn.utils import shuffle


class DeepInput:
    """
    Class for writing training set compatible with deepmd-kit from ASE/VASP output
    (.traj, .db, vasprun.xml, etc.).

    Args:
        atoms_file (str): File containing training set configurations (e.g., .db).

        atoms (ase.Atoms or None): Optional Atoms object to use for assigning atom types.
            First image in atoms_file is used if None specified.

        system_name (str): Descriptive name of atomic system to use as directory name in
            dataset folder.

        type_map (dict or None): Dictionary mapping each atomic symbol to corresponding
            atom type index, optional and alphabetic order if used if None specified.
            E.g., {'Si': 0, 'O': 1}.

        n (int): Max number of images to take from atoms_file. All images are randomly
            shuffled and then n are taken for training.

        path (str): Path to dataset parent folder, makes folder if doesn't already exist.

        append (bool): If True, appends new configurations to current dataset if system_name
            dataset already exists.
    """

    def __init__(self, atoms_file, atoms=None, system_name=None, type_map=None, append=False, n=None, path="./data"):
        self.atoms_file = atoms_file
        if atoms is not None:
            self.atoms = atoms
        self.system_name = system_name
        self.n = n
        self.path = path
        self.type_map = type_map
        self._append = append
        self.set_dataset()
        self.write_input()

    def set_dataset(self):
        positions = []
        energies = []
        forces = []
        box = []
        n = self.n

        if self.atoms_file.endswith(".db"): # TODO: Rework this mess
            with connect(self.atoms_file) as db:
                for row in db.select():
                    if not hasattr(self, "atoms"):
                        self.atoms = row.toatoms() # saving for atom typing
                    self._check_indexing(list(row.numbers))
                    positions.append(row.positions.flatten())
                    forces.append(row.forces.flatten())
                    energies.append(row.energy)
                    box.append(row.cell.flatten())
        else:
            all_atoms = read(self.atoms_file, index=":")
            for atoms in all_atoms:
                if not hasattr(self, "atoms"):
                    self.atoms = atoms.copy() # saving for atom typing
                self._check_indexing(list(atoms.numbers))
                positions.append(atoms.positions.flatten())
                forces.append(atoms.get_forces().flatten())
                energies.append(atoms.get_potential_energy())
                box.append(atoms.cell.array.flatten())

        positions = np.array(positions)
        forces = np.array(forces)
        energies = np.array(energies)
        box = np.array(box)
        positions, energies, forces, box = shuffle(positions, energies, forces, box)

        if n is not None and n < len(energies):
            positions = positions[:n]
            energies = energies[:n]
            forces = forces[:n]
            box = box[:n]

        self.positions = positions
        self.energies = energies
        self.forces = forces
        self.box = box

    def _check_indexing(self, numbers):
        if not hasattr(self, "_ref"):
            self._ref = list(self.atoms.numbers)
            return
        if self._ref != numbers:
            if len(self._ref) != len(numbers):
                err = f"Multiple unique systems detected for {self.atoms_file}."
            else:
                err = f"Inconsistent indexing detected for {self.atoms_file}."
            symbol_map = {v: k for k, v in atomic_numbers.items()}
            ref_syms = [symbol_map[n] for n in self._ref]
            syms = [symbol_map[n] for n in numbers]
            err += f"\nConflicting system indices:\n\n{ref_syms}\n\n{syms}"
            raise Exception(err)

    def write_input(self):
        if self.system_name is None:
            system_name = self.atoms_file.split("/")[-1].split(".")[0]
        else:
            system_name = self.system_name
        data_path = os.path.join(self.path, system_name)
        sets = ["train", "validation", "test"]
        self.paths = {s: os.path.join(data_path, s, "set.000") for s in sets}
        for s, path in self.paths.items():
            os.makedirs(path, exist_ok=True)

        n_train = int(np.ceil(len(self.energies) * 0.80)) # 80% train
        n_val = int(np.ceil(len(self.energies) * 0.90)) # 10% test, 10% val
        if n_val - n_train < 10: # need min of 10 for validation set
            n_train = n_val - 10
            if n_train < 1:
                raise ValueError(f"Need more images in {self.atoms_file}, \
                        {len(self.energies)} entries")

        self.write_types()
        self.write_npy_set("train", ":" + str(n_train))
        self.write_npy_set("validation", str(n_train) + ":" + str(n_val))
        self.write_npy_set("test", str(n_val) + ":")

    def write_npy_set(self, dataset, indices):
        indices = string2index(indices)
        path = self.paths[dataset]
        self._write_npy_file(path, "coord", self.positions[indices])
        self._write_npy_file(path, "force", self.forces[indices])
        self._write_npy_file(path, "energy", self.energies[indices])
        self._write_npy_file(path, "box", self.box[indices])

    def _write_npy_file(self, path, key, vals):
        file_name = os.path.join(path, f"{key}.npy")
        if self._append and os.path.exists(file_name):
            old_vals = np.load(file_name)
            if old_vals.ndim == 2 and old_vals.shape[-1] != vals.shape[-1]:
                # check for matching system sizes if appending to old dataset
                raise ValueError(f"Tried appending to {file_name} but size mismatch found.")
            vals = np.append(old_vals, vals, axis=0)
        np.save(file_name, vals)

    def write_types(self):
        symbols = self.atoms.get_chemical_symbols()
        if self.type_map is None:
            elements = np.unique(symbols)
            type_keys = {e: i for i, e in enumerate(elements)}
        else:
            # inverting type_map, confusing and should probably be reworked
            type_keys = {v: k for k, v in self.type_map.items()}

        types = [type_keys[s] for s in symbols]

        type_paths = [os.path.join(path, "../type.raw") for s, path in self.paths.items()]
        for path in type_paths:
            np.savetxt(path, types, fmt="%.1i", newline=" ")

        self.type_map = {v: k for k, v in type_keys.items()}


class DeepInputs:
    """
    Class for writing training set compatible with deepmd-kit from multiple ASE/VASP outputs
    (.traj, .db, vasprun.xml, etc.).

    Args:
        db_names (str): File containing training set configurations (e.g., .db, .traj).
            TODO: update variable name to something more generic
        atoms (ase.Atoms or None): Optional Atoms object to use for assigning atom types.
            First image in atoms_file is used if None specified.
        system_names (list[str]): Descriptive name of atomic systems to use as directory names in
            dataset folder. Each name should be descriptive (e.g. '00_sodalite_10h2o') and unique
            for each db_names item, i.e., len(system_names) must equal len(db_names)
        type_map (dict or None): Dictionary mapping each atomic symbol to corresponding
            atom type index, optional and alphabetic order if used if None specified.
            E.g., {'Si': 0, 'O': 1}.
        n (int): Max number of images to take from atoms_file. All images are randomly
            shuffled and then n are taken for training.
        path (str): Path to dataset parent folder, makes folder if doesn't already exist.
    """

    def __init__(self,
                 db_names,
                 atoms=None,
                 system_names=None,
                 type_map=None,
                 append=False,
                 n=None,
                 in_json=None,
                 path="./data",
                 ):

        self.path = path
        os.makedirs(self.path, exist_ok=True)
        if atoms is None:
            atoms = self.get_atoms(db_names)
        elif not isinstance(atoms, list):
            atoms = [atoms]
        if system_names is None:
            system_names = [None] * len(db_names)
        elif isinstance(system_names, str):
            system_names = [system_names]
        if isinstance(db_names, str):
            db_names = [db_names]
        if type_map is None:
            type_map = self.get_type_map(atoms)
        if in_json is None:
            default_path = os.path.abspath(os.path.dirname(__file__))
            in_json = os.path.join(default_path, "in.json")

        self._check_names(db_names, system_names)

        self.type_map = type_map
        self._json_file = in_json

        for db, a, sys in zip(db_names, atoms, system_names):
            dpi = DeepInput(db, a, sys, self.type_map, append=append, n=n, path=path)

        self.set_json()
        self.update_json()
        self.write_json()

    def set_json(self):
        with open(self._json_file, "r") as file:
            self.input_json = json.loads(file.read())

    def update_json(self):
        self.set_systems()

        types = [self.type_map[i] for i in range(len(self.type_map))]
        self.input_json["model"]["type_map"] = types

    def set_systems(self):
        systems = [s for s in glob.glob(f"{self.path}/*") if os.path.isdir(s)]
        systems.sort()

        def get_paths(key):
            return [os.path.join(s, key) for s in systems]

        self.input_json["training"]["training_data"]["systems"] = get_paths("train")
        self.input_json["training"]["validation_data"]["systems"] = get_paths("validation")

    def write_json(self):
        json_str = json.dumps(self.input_json, indent=4)
        with open("in.json", "w") as file:
            file.write(json_str)

    @staticmethod
    def _check_names(input_files, system_names):
        n_in, n_sys, n_unique = len(input_files), len(system_names), len(np.unique(system_names))
        check1 = n_in != n_sys
        check2 = None not in system_names and n_unique != n_sys
        if check1 or check2:
            if check1:
                err = f"{n_sys} system_names provided for {n_in} input files"
            else:
                err = "Duplicate names detected in system_names, need all unique"
            raise ValueError(err)

    def get_atoms(self, db_names):
        atoms = []
        for dbn in db_names:
            if dbn.endswith(".db"):
                with connect(dbn) as db:
                    for row in db.select():
                        atoms.append(row.toatoms())
                        break
            else:
                atoms.append(read(dbn, index="0"))
        return atoms

    def get_type_map(self, atoms):
        """
        Check Atoms object from all systems for unique symbols, and then
        assigns type_map in alphabetical order.

        Args:
            atoms (list[ase.Atoms]): List of example Atoms object from all systems
                that will be used for training (e.g. the first image from all vasprun inputs).

        Returns:
            type_map (dict): Dictionary mapping each atom type index to atomic symbol in
                alphabetical order.
        """
        tm_path = os.path.join(self.path, "type_map.json")
        if "type_map.json" in os.listdir(self.path):
            print(f"READING {tm_path}")
            with open(tm_path, "r") as file:
                type_map = json.loads(file.read())
            type_map = {int(i): s for i, s in type_map.items()}
        else:
            symbols = []
            for a in atoms:
                symbols.extend(np.unique(a.get_chemical_symbols()))

            symbols = np.unique(symbols)
            type_map = dict(enumerate(symbols))
            print(f"WRITING TYPE MAP TO {tm_path}")
            with open(tm_path, "w") as file:
                file.write(json.dumps(type_map, indent=2))

        print("TYPES:")
        for i, t in type_map.items():
            print(f"\t{i}\t{t}")
        print("If unhappy with above type ordering, edit type_map.json to your liking and rerun!")

        return type_map



import re
from textwrap import dedent
import numpy as np
from ase.data import atomic_masses, atomic_numbers
from ase.geometry.analysis import Analysis
from ase.calculators.lammpslib import convert_cell


class LammpsInput:
    """
    Creates lammps input files from ASE Atoms object. Works by setting as str attributes
    all neccessary lammps input information and commands, which are then inserted into
    the appropriate Template subclass.


    Args:
        atoms (ase.Atoms): structure used for lammps calculation

        type_dict (dict): map atom type index to element symbol
                   (Element_tag to specify multiple types of same element)
                   e.g., {1: 'O', 2: 'Si', 3: 'O_h2o', 4: 'H_h2o'}

        charges (array-like): List or array of charges (float) on each atom.
            Must have len(charges) == len(atoms).

        bonds (bool): Define all bonds in structure data file if True.
        angles (bool): Define all angles in structure data file if True.
        dihedrals (bool): Define all dihedral angles in structure data file if True.
        impropers (bool): Define impropers in structure data file if True.

        name (str, optional): Name used for creating input files (name.data, in.name).

        atom_style (str, optional): lammps style for atomic positions, groups, etc.
        pair_style (str, optional): lammps pair_style command.
        bond_style (str, optional): lammps bond_style command.
        angle_style (str, optional): lammps angle_style command.
        kspace_style (str, optional): lammps kspace_style command.

        groups (array-like, optional): List of groups (int) for each atom.
            Must have len(groups) == len(atoms).

        pair_coeff (list[str] or str, optional): lammps pair_coeff line(s) to define ij interactions
    """

    def __init__(self, atoms, type_dict,
                 charges=None,
                 bonds=False,
                 angles=False,
                 dihedrals=False,
                 impropers=False,
                 name="atoms",
                 atom_style="full",
                 pair_style=None,
                 bond_style=None,
                 angle_style=None,
                 kspace_style=None,
                 pair_coeff=None,
                 bond_coeff=None,
                 angle_coeff=None,
                 groups=None,
                 ):

        self.atoms = atoms
        self.set_lmp_cmd("atom_style", atom_style)
        self.set_lmp_cmd("pair_style", pair_style)
        self.set_lmp_cmd("bond_style", bond_style)
        self.set_lmp_cmd("angle_style", angle_style)
        self.set_lmp_cmd("kspace_style", kspace_style)
        self.set_lmp_cmd("pair_coeff", pair_coeff)
        self.set_lmp_cmd("bond_coeff", bond_coeff)
        self.set_lmp_cmd("angle_coeff", angle_coeff)
        self.atom_style_hint = self.atom_style.split()[-1]

        if charges is None:
            charges = atoms.get_initial_charges()
        self.charges = charges
        self.groups = groups
        self.type_dict = type_dict
        self._bonds = bonds
        self._angles = angles
        self._dihedrals = dihedrals
        self._impropers = impropers
        self.name = name

    def set_lmp_cmd(self, command, text):
        """
        Set lammps commands that follow the pattern: command args

        Args:
            command (str): Name of lammps command, e.g., pair_style.
            text (list[str] or str): Arguments to use with command. If text doesn't
                lead with command, then command is prepended to text.
        """
        if isinstance(text, list):
            lines = [self.prepend_command(command, l) for l in text]
            new_text = "\n".join(lines)
        else:
            new_text = self.prepend_command(command, text)
        setattr(self, command, new_text or "")

    @staticmethod
    def prepend_command(command, text):
        """
        Prepend command to corresponding arguments str.

        Args:
            command (str): Name of lammps command to prepend to text.
            text (str): lammps arg str corresponding to command.

        Returns:
            str: Full lammps line with command prepended to text.

        Examples:
            self.prepend_command('kspace_style', 'pppm 1e-5')
            'kspace_style pppm 1e-5'
        """
        if text and not text.startswith(command):
            text = f"{command} {text}"
        return text

    def write(self, atoms=None, charges=None):
        """
        Writes lammps data file and input file for atoms.

        Args:
            atoms (ase.Atoms or None): New atoms object used to write data file
                (or self.atoms if atoms is None).
            charges (array-like or None): Optional charges corresponding to atoms.
        """
        self.write_atoms(atoms=atoms, charges=charges)
        self.write_infile()

    def write_atoms(self, atoms=None, charges=None):
        """
        Write lammps data file from atoms. Writes unit cell info, atom types,
        atomic positions, and bonds, angles, dihedrals, impropers if specified in __init__().

        Args:
            atoms (ase.Atoms or None): New atoms object used to write data file
                (or self.atoms if atoms is None).
            charges (array-like or None): Optional charges corresponding to atoms.
        """
        atoms = atoms if atoms is not None else self.atoms
        charges = charges if charges is not None else self.charges
        self.natoms = len(atoms)
        self.write_cell()
        self.write_types()
        self.write_coords()
        self.write_constraints(atoms)

        for key in ["bonds", "angles", "dihedrals", "impropers"]:
            self.write_geometry(key)

        datatemp = DataTemplate(f"data.{self.name}", self)
        datatemp.write()

    def write_cell(self):
        """Transform and write self.atoms' ``ase.cell.Cell`` to lammps data file."""
        cell = convert_cell(self.atoms.cell)[0]
        self.x = f"0 {cell[0, 0]}"
        self.y = f"0 {cell[1, 1]}"
        self.z = f"0 {cell[2, 2]}"
        xy, xz, yz = cell[0, 1], cell[0, 2], cell[1, 2]
        self.xyxzyz = f"{xy} {xz} {yz}"

    def write_types(self):
        """Write Types section in lammps data file (atomic masses for each atom type)."""
        self.ntypes = len(self.type_dict.keys())

        types_str = ""
        for k, v in self.type_dict.items():
            sym = v.split("_")[0]
            mass = atomic_masses[atomic_numbers[sym]]
            types_str += f"{k} {mass} # {v}\n"

        self.types = types_str

    def write_coords(self):
        """Write Coords section (atomic positions) in lammps datafile."""
        coords_str = ""
        zeros = np.zeros(self.natoms, dtype=int)
        groups = self.groups if self.groups is not None else zeros
        for a, g, q, (x, y, z) in zip(self.atoms,
                                      groups,
                                      self.charges,
                                      self.atoms.positions):

            if not self.atom_style or self.atom_style == "atom_style full":
                coords_str += f" {a.index + 1} {g} {a.tag} {q} {x} {y} {z}\n"
            else:
                err = f"atom_style {self.atom_style} not supported, harass me for it, "\
                    "but I can't imagine why you'd need it for deepmd."
                raise NotImplementedError(err)
                #coords_str += f" {a.index + 1} {a.tag} {x} {y} {z}\n"

        self.coords = coords_str

    def write_constraints(self, atoms):
        if not atoms.constraints:
            self.constraints = ""
            return
        indices = atoms.constraints[0].get_indices().copy()
        indices += np.ones(len(indices), dtype="int64")
        constrained = " ".join(list(map(str, indices)))
        self.constraints = f"group constrained id {constrained}"
        self.constraints += "\ngroup unconstrained subtract all constrained"
        self.constraints += "\n\nfix constraints constrained setforce 0.0 0.0 0.0"

    def write_geometry(self, key):
        """
        Generalized method to write bonds, angles, maybe dihedrals (untested) to lammps data file.
        Uses ``ase.geometry.Analysis`` to generate list of bonds, angles, etc.

        Args:
            key (str): Geometry component to write ('bonds', 'angles', 'dihedrals', 'impropers').
        """
        n = 0
        n_types = 0
        text = ""
        if getattr(self, f"_{key}"): # e.g. if _bonds, write bonds
            if key in ["dihedrals", "impropers"]:
                raise NotImplementedError(f"{key} not implemented, harass me if you need it")

            # define order of indices when writing key
            # (e.g. angles need central atom in middle, 1-0-2)
            orders = {"bonds": "01", "angles": "102", "dihedrals": "0123"}
            if not hasattr(self, "anal"): # only calc once
                self.anal = Analysis(self.atoms)

            def t_str(order, *t):
                tsorted = [str(sorted(t)[int(i)]) for i in order]
                return "-".join(tsorted)

            tags = self.atoms.get_tags()
            types = {}
            unique = getattr(self.anal, f"unique_{key}")[0]
            for i, groups in enumerate(unique):
                for neighbors in groups:
                    if isinstance(neighbors, np.int32):
                        neighbors = [neighbors] # bonds only have single (int) neighbor
                    typ = t_str(orders[key], tags[i], *[tags[n] for n in neighbors])

                    if typ in types:
                        types[typ].append((i, *neighbors))
                    else:
                        types[typ] = [(i, *neighbors)]

            n_types = len(types)
            text = key.capitalize() + "\n\n"
            #for i, (typ, groups) in enumerate(types.items()):
            for i, typ in enumerate(sorted(types.keys())):
                for group in types[typ]:
                    g_str = "\t".join([str(g + 1) for g in group])
                    text += f" {n+1}\t{i+1}\t{g_str}\n"
                    n += 1

        setattr(self, key, text)
        setattr(self, f"n{key}", n)
        setattr(self, f"n{key[0]}types", n_types)

    def write_infile(self):
        """
        Write lammps input file containing all general setup commands
        (e.g. define units, read data file, etc.).
        """
        if not self.pair_coeff:
            self.pair_coeff = "pair_coeff * *\n"

        if self.groups is not None:
            raise NotImplementedError("Harass me if you need it")

        intemp = InputTemplate(f"in.{self.name}", self)
        intemp.write()

    def _write_geometry(self, key):
        """
        Deprecated, use new write_geometry. Generalized method to write geometry components
        (bonds, angles, dihedrals, or impropers) to lammps data file and set quantities
        (num bonds, bond types, etc.).

        Args:
            key (str): geometry component to write
                ('bonds', 'angles', 'dihderals', or 'impropers').
        """
        if getattr(self, f"_{key}"): # e.g.  if self._bonds:
            writer = getattr(self, f"write_{key}") # self.write_bonds()
            writer()
        else:
            setattr(self, key, "")
            setattr(self, f"n{key}", 0)
            setattr(self, f"n{key[0]}types", 0)

    def write_bonds(self):
        """
        (Deprecated, may restore) Write all bonds in self.atoms to lammps datafile.
        """
        self.anal = Analysis(self.atoms)
        tags = self.atoms.get_tags()
        bond_types = {}

        def b_str(t1, t2):
            return f"{t1}-{t2}" if t1 < t2 else f"{t2}-{t1}"

        for i, bonds in enumerate(self.anal.unique_bonds[0]):
            for bond in bonds:
                b_type = b_str(tags[i], tags[bond])
                if b_type in bond_types:
                    bond_types[b_type].append((i, bond))
                else:
                    bond_types[b_type] = [(i, bond)]

        bonds_str = "Bonds\n\n"
        n_bond = 0
        for i, (b_type, bonds) in enumerate(bond_types.items()):
            for bond in bonds:
                bonds_str += f"\t{n_bond+1}\t{i+1}\t{bond[0]+1}\t{bond[1]+1}\n"
                n_bond += 1

        self.nbonds = n_bond
        self.nbtypes = len(bond_types.keys())
        self.bonds = bonds_str

    def write_angles(self):
        """
        (Deprecated, may restore) Write all angles in self.atoms to lammps datafile.
        """
        if not hasattr(self, "anal"): # don't recalc if already done in write_bonds()
            self.anal = Analysis(self.atoms)
        tags = self.atoms.get_tags()
        angle_types = {}

        def a_str(t1, t2, t3):
            t = sorted([t1, t2, t3])
            return f"{t[1]}-{t[0]}-{t[2]}"

        for i, angles in enumerate(self.anal.unique_angles[0]):
            for angle in angles:
                a_type = a_str(tags[i], tags[angle[0]], tags[angle[1]])
                if a_type in angle_types:
                    angle_types[a_type].append((i, angle))
                else:
                    angle_types[a_type] = [(i, angle)]

        raise NotImplementedError("Too lazy to add angles, harass me if you need it")


    def write_dihedrals(self):
        """Deprecated before it was even written :("""
        raise NotImplementedError("Too lazy to add dihedrals, harass me if you need it")

    def write_impropers(self):
        """Deprecated before it was even written :("""
        raise NotImplementedError("Harass me if you need this.")


class MolInput:
    """
    Class for writing simple molecule files in lammps for deepmd.

    TODO: add bonds, charges, etc. for future applications

    Args:
        atoms (ase.Atoms): Atoms object of molecule to write.

        type_dict (dict): map atom type index to element symbol
                   (Element_tag to specify multiple types of same element)
                   e.g., {1: 'O', 2: 'Si', 3: 'O_h2o', 4: 'H_h2o'}

        name (str): Name of molecule for saving file (mol.name)
    """
    def __init__(self, atoms, type_dict, name="mol"):
        self.atoms = atoms
        self.type_dict = type_dict
        self.name = name

    def write(self):
        self.write_atoms()
        moltemp = MolTemplate(f"mol.{self.name}", self)
        moltemp.write()

    def write_atoms(self):
        self.natoms = len(self.atoms)
        coords = "Coords\n\n"
        types = "Types\n\n"

        for a, (x, y, z) in zip(self.atoms, self.atoms.positions):
            coords += f"{a.index + 1}\t\t{x}\t{y}\t{z}\n"
            types += f"{a.index + 1}\t\t{a.tag}\n"

        self.coords = coords
        self.types = types


class Template:
    """
    Base class for writing lammps data and input files. Replaces all <attr> sections
    in subclass self.text with corresponding LammpsInput instance attributes that contain
    strings for each section.

    Args:
        file_name (str): Name of file to write, generally name.data or in.name.
        lammps_input (LammpsInput instance): LammpsInput object with str attributes used
            to fill in self.text with appropriate commands / values.
    """
    def __init__(self, file_name, lammps_input):
        self.file_name = file_name
        self.input = lammps_input

    def write(self):
        """
        Write lammps file to current directory.
        """
        self.fill()
        self.trim()
        with open(self.file_name, "w") as file:
            file.write(self.text)

    def fill(self):
        """
        Replaces <attr> sections in self.text with corresponding LammpsInput object attribute text.
        """
        pattern = r"<\w+>"
        matches = re.findall(pattern, self.text)
        for match in matches:
            key = match[1:-1]
            attr = getattr(self.input, key)
            attr = str(attr) if attr is not None else ""
            self.text = self.text.replace(match, attr)

    def trim(self):
        """
        Remove unneccessary blank lines from self.text caused by inserting empty strings
        as commands.
        """
        while "\n\n\n" in self.text:
            self.text = self.text.replace("\n\n\n", "\n\n")


# XXX: Don't really have a need for subclasses anymore, but it works
class DataTemplate(Template):
    text = dedent('''\
        LAMMPS data file

          <natoms> atoms
          <nbonds> bonds
          <nangles> angles
          <ndihedrals> dihedrals
          <nimpropers> impropers

          <ntypes> atom types
          <nbtypes> bond types
          <natypes> angle types
          <ndtypes> dihedral types
          <nitypes> improper types

          <x> xlo xhi
          <y> ylo yhi
          <z> zlo zhi
          <xyxzyz> xy xz yz

        Masses

        <types>

        Atoms # <atom_style_hint>

        <coords>

        <bonds>

        <angles>

        <dihedrals>

        <impropers>
        ''')


class InputTemplate(Template):
    text = dedent('''\
        units\t\tmetal
        boundary\tp p p
        <atom_style>
        atom_modify\tmap array

        <pair_style>
        <bond_style>
        <angle_style>

        box tilt\tlarge
        read_data\tdata.<name>

        <constraints>

        <pair_coeff>

        <bond_coeff>

        <angle_coeff>

        <kspace_style>

        <groups>
        thermo_style custom step temp etotal pe press pxx pyy pzz pxy pxz pyz lx ly lz vol
        ''')


class MolTemplate(Template):
    text = dedent('''
        <natoms> atoms
        0 bonds
        0 angles

        <coords>

        <types>
        ''')



from dptools.cli import BaseCLI
from dptools.utils import Converter

class CLI(BaseCLI):
    """
    Convert between structure ASE/VASP/LAMMPS file types.

    :doc:`Complete documentation here<../commands/convert>`

    Examples:

    .. code-block:: console

        $ dptools convert thing.cif thing.traj
        $ dptools convert md_000/vasprun.xml md_001/vasprun.xml full_md.traj
        $ dptools convert md_???/vasprun.xml full_md.traj # equivalent to above
        $ dptools convert -i ::10 vasprun.xml condensed_traj.traj
    """

    help_info = "Convert between structure file types (e.g., .xml to .db)"
    def add_args(self):
        self.parser.add_argument(
            "inputs",
            nargs="+",
            metavar="input",
            help="Input files (with extensions) to convert. Multiple inputs are concatenated into output",
        )
        self.parser.add_argument(
            "output",
            nargs=1,
            help="Output file name to write conversion to (with extension)",
        )
        self.parser.add_argument("-i", "--indices", type=str, default=":",
                help="Indices of input files to read. E.g., :10, -3:, :100:5")
        self.parser.add_argument("-r", "--repeat", type=str, default=None,
                help="Repeat each input before writing to output (e.g., 222)")

    def main(self, args):
        converter = Converter(args.inputs, args.output[0], args.indices)
        self.get_kwargs(args.inputs)
        converter.read(**self.kwargs)
        if args.repeat:
            if len(args.repeat) != 3:
                raise ValueError("Need 3 numbers for repeat arg - e.g., 222")
            repeat = [int(i) for i in args.repeat]
            for i, atoms in enumerate(converter.atoms):
                converter.atoms[i] = atoms.repeat(repeat)
        converter.convert()

    def get_kwargs(self, inputs):
        self.kwargs = {}
        if inputs[0].endswith(".dump"):
            from dptools.env import get_dpfaults
            graph, type_map = get_dpfaults()
            self.kwargs.setdefault("type_map", type_map)



import glob
import re
from os.path import dirname, join, abspath

from dptools.cli import BaseCLI
from dptools.env import get_env, load
from dptools.utils import str2typemap


class CLI(BaseCLI):
    """
    Display information on set DP models and Slurm parameters.

    :doc:`Complete documentation here<../commands/info>`

    Examples:

    .. code-block:: console

        $ dptools info # display all set environments
        $ dptools info -m water # display specific labelled environment
    """
    help_info = "Show loaded DP models and sbatch parameters"

    def add_args(self):
        self.parser.add_argument("-m", "--model-label", type=str, default=all,
                help="Label of specific model to use (see dptools set -h)")

    def main(self, args):
        self.keys = [
                "DPTOOLS_MODEL",
                 "DPTOOLS_TYPE_MAP",
                 "DPTOOLS_MODEL2",
                 "DPTOOLS_MODEL3",
                 "DPTOOLS_MODEL4",
                 "SBATCH_COMMENT"
             ]
        basedir = abspath(join(dirname(__file__), ".."))
        envs = sorted(glob.glob(f"{basedir}/.env*"))
        if args.model_label is not all:
            envs = [e for e in envs if e.endswith(args.model_label)]
        self.info = {}
        for env in envs:
            self.set_info(env)

        self.summarize()

    @staticmethod
    def get_env_name(env_file):
        """
        Get corresponding name for .env.* file, e.g. water from .env.water.

        Args:
            env_file (str): path to .env file to get corresponding label (name).

        Returns:
            label (str): name of environment that is appended to .env file.
        """
        pattern = ".env.[a-zA-Z0-9]+"
        if re.search(pattern, env_file):
            label = env_file.split(".")[-1]
        else:
            label = ""
        return label

    def set_info(self, env_file):
        """
        Get and set all env key-value info.

        Args:
            env_file (str): path to .env file to get and set info for.
        """
        label = self.get_env_name(env_file)
        if label:
            load(label)
        else:
            label = "default"
        self.info[label] = get_env()

    def summarize(self):
        """Print out semi-formatted env info."""
        if not self.info:
            print("No info to display, try setting a model using:",
                    "\t dptools set /path/to/graph.pb")
        for env, vals in self.info.items():
            print("-" * 64)
            print(f"{env} env")
            print()
            self._print_kv(vals, "DPTOOLS_MODEL")
            self._print_kv(vals, "DPTOOLS_MODEL2")
            self._print_kv(vals, "DPTOOLS_MODEL3")
            self._print_kv(vals, "DPTOOLS_MODEL4")
            print()
            self._print_kv(vals, "DPTOOLS_TYPE_MAP")
            print()
            self._print_kv(vals, "SBATCH_COMMENT", fmt="{key}:\n{val}")
            print("-" * 64)
            print()

    @staticmethod
    def _print_kv(thing, key, fmt="{key}={val}"):
        val = thing.get(key)
        if "TYPE_MAP" not in key:
            if val:
                print(fmt.format(key=key, val=val))
        else:
            print("TYPE MAP:")
            if val:
                type_map = str2typemap(val)
                for k, v in type_map.items():
                    print(f"{k}\t{v}")
            else:
                print("NO MODEL SET")



import os
import numpy as np
from ase import units
from ase.io import read, write

from dptools.simulate.calculator import DeepMD
from dptools.simulate.lammps_io import MolInput
from dptools.utils import read_dump, columnize
from dptools.utils import get_seed as seed


class Simulation:
    """
    Base class for setting up, running, and processing specific simulations in LAMMPS
    using DP models for energy and force evaluations.

    Args:
        atoms (str, ase.Atoms, or list[ase.Atoms]): Input structure(s) to run
            simulation with. ase.Atoms object or str with path to .traj (or similar) file.

        graph (str): Path to deepmd model .pb file

        type_map (dict): Dictionary that maps elemental symbol keys to type index values.

        path (str): Path to directory to write results to if dir other than $PWD desired.

        **kwargs: Simulation specific kwargs to pass to setup and get_commands methods.
    """

    def __init__(self, atoms, graph, type_map, file_out="atoms.traj", path="./", **kwargs):
        if isinstance(atoms, str):
            atoms = [read(atoms)]
        elif not isinstance(atoms, list):
            atoms = [atoms]
        self.atoms = atoms

        self.graph = graph
        self.type_map = type_map
        self.path = path
        self.file_out = os.path.join(path, file_out)
        self.setup(**kwargs)

    def setup(self, **kwargs):
        """
        Simulation specific method to do any needed setup before running (e.g., cell deformations).
        Must call self.get_commands() method.
        """
        self.commands = self.get_commands(**kwargs)

    def run(self, process=True, commands=None, file_out=None):
        """
        Runs simulation after setting everything up.

        Args:
            process (bool): If True calls self.process() method after running.

            commands (list[str], optional): List of lammps commands if different from self.commands.
                Rarely used, should set self.commands in self.setup() generally.

            file_out (optional): Name of output file if different from self.file_out for
                whatever reason. Also rarely used.
        """
        commands = commands if commands else self.commands
        for atoms in self.atoms:
            calc = DeepMD(self.graph, type_map=self.type_map, run_command=commands, verbose=True)
            atoms.calc = calc
            atoms.get_potential_energy()
        if process:
            self.process(file_out=file_out)

    def process(self, file_out=None):
        """Simulation specific method to process and write results after calculation."""
        file_out = file_out if file_out else self.file_out
        write(file_out, self.atoms)

    def pre_opt(self, nsw, cell=False, ftol=0.01):
        """
        Run quick optimization on structure before running full simulation.

        Args:
            nsw (int): Max number of optimization steps to run.
            cell (bool): Optimize unit cell if True, else only positions.
            ftol (float): Max force convergence tolerance criterion.
        """
        Opts = {0: Opt, 1: CellOpt}
        commands = Opts[cell].get_commands(self, nsw=nsw, ftol=ftol)

        Simulation.run(self, process=False, commands=commands)

        file_out = os.path.join(self.path, "pre_opt.traj")
        write(file_out, self.atoms)

    def _warn_unused(self, **kwargs):
        for k, v in kwargs.items():
            print(f"WARNING: {k}={v} unused for calculation type {self.calc_type}")

    def write_array(self, data):
        name = f"data.{self.calc_type}.npy"
        np.save(os.path.join(self.path, name), data)


class NewSimulation(Simulation):
    """
    Template/example for making new simulation types in lammps.

    Args:
        setup_arg1: Argument from params.yaml that is passed to setup method
        setup_arg2: Argument from params.yaml that is passed to setup method
        param_arg1: Argument from params.yaml that is passed to get_commands method
        param_arg2: Argument from params.yaml that is passed to get_commands method
    """

    def setup(self, setup_arg1=None, setup_arg2=None, **kwargs):
        """
        Simulation specific method to do any needed setup before running (e.g., cell deformations).
        Must call self.get_commands() method.
        """

        # do stuff with setup args here

        # then set commands
        self.commands = self.get_commands(**kwargs)

    def get_commands(self, **kwargs):
        self._warn_unused(**kwargs) # warn about any unused kwargs from params.yaml

        # set all simulation lammps commands (besides initial setup, e.g. unit styles)
        commands = ["run 0"]
        return commands

    def run(self, file_out=None):
        """
        Rarely need subclass run method. Test without it and if Simulation.run fails
        for whatever reason, write your own run method.
        """
        pass

    def process(self, file_out=None):
        """
        Only needed if additional proceessing from lammps output is necessary.
        Also useful for printing results summaries.
        """
        pass


class SPE(Simulation):
    """Simple single point energy calculation, not much to it."""
    calc_type = "spe"

    def get_commands(self, **kwargs):
        self._warn_unused(**kwargs)
        commands = ["run 0"]
        return commands


class Opt(Simulation):
    calc_type = "opt"

    def get_commands(self, nsw=1000, ftol=1e-2, etol=0.0, disp_freq=10, **kwargs):
        self._warn_unused(**kwargs)
        commands = [
            f"thermo {disp_freq}",
             "min_modify norm max",
            f"minimize {etol} {ftol} {nsw} {nsw * 10}",
            ]
        return commands


class CellOpt(Simulation):
    calc_type = "cellopt"

    def get_commands(self, nsw=1000, ftol=1e-2, etol=0.0, opt_type="aniso", Ptarget=0.0, disp_freq=10, **kwargs):
        self._warn_unused(**kwargs)
        commands = [
                f"thermo {disp_freq}",
                 "min_modify norm max",
                f"fix cellopt all box/relax {opt_type} {Ptarget}",
                f"minimize {etol} {ftol} {nsw} {nsw * 10}",
                 "unfix cellopt",
                ]
        return commands


class NVT(Simulation):
    calc_type = "nvt-md"

    def setup(self, pre_opt=False, **kwargs):
        if pre_opt:
            self.pre_opt(200)

        # group 'unconstrained' created by lammps_io.LammpsInput if constraints present
        # need to specify only unconstrained atoms for velocity command
        self._unconstrained = "unconstrained" if self.atoms[0].constraints else "all"

        self.commands = self.get_commands(**kwargs)

    def get_commands(self, steps=10000, timestep=0.5, Ti=298.0, Tf=298.0, equil_steps=1000, write_freq=100, disp_freq=100, **kwargs):
        self._warn_unused(**kwargs)
        timestep = timestep * 1e-3 # convert to ps for lammps
        commands = [
            f"thermo {disp_freq}",
            f"variable\tdt\tequal\t0.5e-3",
            "variable\ttdamp\tequal 100*${dt}",
            "run_style verlet",
            "timestep ${dt}",
            # XXX: Add customizable velocity keywords as args?
            f"velocity {self._unconstrained} create {Ti} {seed()} rot yes mom yes dist gaussian",
            f"fix equil all nvt temp {Ti} {Ti} ${{tdamp}}",
            f"run {equil_steps}",
             "unfix equil",
            f"fix nvt_prod all nvt temp {Ti} {Tf} ${{tdamp}}",
            f"dump 1 all custom {write_freq} nvt.dump id type x y z",
            f"run {steps}"
            ]
        return commands

    def process(self, file_out=None):
        atoms = read_dump("nvt.dump", self.type_map)
        write(self.file_out, atoms)


class NPT(Simulation):
    calc_type = "npt-md"

    def setup(self, pre_opt=False, **kwargs):
        if pre_opt:
            self.pre_opt(200, cell=True)

        # group 'unconstrained' created by lammps_io.LammpsInput if constraints present
        # need to specify only unconstrained atoms for velocity command
        self._unconstrained = "unconstrained" if self.atoms[0].constraints else "all"

        self.commands = self.get_commands(**kwargs)

    def get_commands(self, steps=10000, timestep=0.5, Pi=0.0, Pf=0.0, Ti=298.0, Tf=298.0, equil_steps=1000, write_freq=100, disp_freq=100, **kwargs):
        self._warn_unused(**kwargs)
        timestep = timestep * 1e-3 # convert to ps for lammps
        commands = [
            f"thermo {disp_freq}",
            f"variable\tdt\tequal\t0.5e-3",
            "variable\tpdamp\tequal 1000*${dt}",
            "variable\ttdamp\tequal 100*${dt}",
            "run_style verlet",
            "timestep ${dt}",
            # XXX: Add customizable velocity keywords as args?
            f"velocity {self._unconstrained} create {Ti} {seed()} rot yes mom yes dist gaussian",
            f"fix equil all npt temp {Ti} {Ti} ${{tdamp}} tri {Pi} {Pi} ${{pdamp}}",
            f"run {equil_steps}",
             "unfix equil",
            f"fix npt_prod all npt temp {Ti} {Tf} ${{tdamp}} tri {Pi} {Pf} ${{pdamp}}",
            f"dump 1 all custom {write_freq} npt.dump id type x y z",
            f"run {steps}"
            ]
        return commands

    def process(self, file_out=None):
        atoms = read_dump("npt.dump", self.type_map)
        write(self.file_out, atoms)


class EOS(Simulation):
    calc_type = "eos"

    def setup(self, N=5, lo=0.96, hi=1.04, pre_opt=True, **kwargs):
        if pre_opt:
            self.pre_opt(200, cell=True)

        self.set_volumes(lo, hi, N)

        self.commands = self.get_commands(**kwargs)

    def get_commands(self, nsw=300, ftol=1e-3, etol=0.0, disp_freq=10, **kwargs):
        self._warn_unused(**kwargs)

        # only need to run standard optimizations on each cell volume
        commands = Opt.get_commands(self, nsw=nsw, ftol=ftol, etol=etol, disp_freq=disp_freq)
        return commands

    def set_volumes(self, lo, hi, N):
        atoms, = self.atoms.copy()
        self.atoms = []
        for v in np.linspace(lo, hi, N):
            new_atoms = atoms.copy()
            new_cell = atoms.cell.array * v ** (1 / 3)
            new_atoms.set_cell(new_cell, scale_atoms=True)
            self.atoms.append(new_atoms)

    def process(self, file_out=None):
        from ase.eos import EquationOfState
        volumes = [a.get_volume() for a in self.atoms]
        energies = [a.get_potential_energy() for a in self.atoms]

        # TODO: Add optional arg to change eos type
        eos = EquationOfState(volumes, energies, eos='birchmurnaghan')
        try:
            v0, e0, B = eos.fit()
            bulk_mod = B / units.kJ * 1.0e24 # [GPa]
            print(f"BULK MODULUS: {bulk_mod:.3f} GPa")
        except RuntimeError:
            print("Bad EOS fit, can not determined bulk modulus")
            print("Check energy versus volume data in data.eos.npy")

        write(self.file_out, self.atoms)
        eos_data = columnize(volumes, energies)
        self.write_array(eos_data)


class Vib(Simulation):
    calc_type = "vib"

    def setup(self, pre_opt=True, **kwargs):
        if pre_opt:
            # tight ftol convergence for vib to ensure minimum is reached
            self.pre_opt(200, ftol=0.001)

        self.commands = self.get_commands(**kwargs)

    def set_displacements(self, nfree, delta):
        # NOTE: deprecated, way more efficient to just use
        #       lammps built in dynamical_matrix command
        if nfree not in [2, 4]:
            raise ValueError("Only supports nfree = 2 or 4")
        atoms, = self.atoms.copy()

        displacements = [-1, 1, -2, 2]
        n_displacements = 3 * nfree * len(atoms)

        for a in atoms:
            for j in range(3): # xyz directions
                for i in range(nfree):
                    d = displacements[i] * delta # magnitude of displacement to translate each atom
                    new_atoms = atoms.copy()
                    new_atoms[a.index].position[j] += d
                    self.atoms.append(new_atoms)

    def get_commands(self, delta=0.015, **kwargs):
        self._warn_unused(**kwargs)
        # calculate dynamical matrix / mass weighted hessian
        commands = [f"dynamical_matrix all eskm {delta} file dynmat.dat"]
        return commands

    def run(self):
        atoms, = self.atoms
        calc = DeepMD(self.graph,
                      type_map=self.type_map,
                      run_command=self.commands,
                      verbose=True,
                      )

        # Need to not update results or lammps crashes, and no reason to write
        # results to Atoms object anyway since we only care about dynmat.dat file
        calc.calculate(atoms, update=False)
        self.process()

    def process(self):
        dyn_mat = np.loadtxt("dynmat.dat")
        n = len(self.atoms[0])
        # reshape lammps dynammical matrix output to square (3n x 3n)
        dyn_mat = dyn_mat.reshape(3 * n, 3 * n)

        eig_vals, eig_vecs = np.linalg.eig(dyn_mat)
        eig_vals = np.flip(np.sort(eig_vals))
        imag_filt = np.array([-1 if e < 0 else 1 for e in eig_vals])

        # speed of light [m/s] https://physics.nist.gov/cgi-bin/cuu/Value?c
        c = 299792458.0
        conversion = 1e12 / (c * 100) # Hz --> cm-1
        freq = np.sqrt(np.abs(eig_vals)) / (2 * np.pi) * conversion
        freq *= imag_filt # save imaginary frequencies as negative

        self.write_array(freq)

        for j, f in enumerate(freq):
            i = "i" if f < 0 else ""
            print(f"Mode {j}: {abs(f):.2f}{i}")


class GCMC(Simulation):
    """Grand-canonical monte carlo simulation."""
    calc_type = "gcmc"

    def setup(self, molecule="H2O", pre_opt=False, pre_opt_mol=False, **kwargs):
        if os.path.isfile(molecule):
            mol = read(molecule)
        else:
            from ase import build
            mol = build.molecule(molecule)
        if mol.cell.sum() == 0.0: # need cell or lammps crashes
            mol.cell = self.atoms[0].cell.copy()
            mol.center()

        mol_tags = [self.type_map[m.symbol] + 1 for m in mol] # tag types for lammps mol writer
        mol.set_tags(mol_tags)

        self._mol = mol.get_chemical_formula().lower() # save for naming files

        if pre_opt_mol:
            optmol = Opt(mol,
                         self.graph,
                         self.type_map,
                         file_out=f"{self._mol}.traj",
                         path=self.path,
                         nsw=200,
                         )
            optmol.run()

        molin = MolInput(mol, self.type_map, name=self._mol)
        molin.write() # write lammps molecule file

        if pre_opt:
            self.pre_opt(200, cell=True)

        self.commands = self.get_commands(**kwargs)

    def get_commands(self, steps=100, n_ex=10, n_mc=10, T=298.0, P=1.0, dmax=1.0, equil_steps=0, write_freq=1, disp_freq=5, **kwargs):
        self._warn_unused(**kwargs)

        commands = [
                "compute_modify  thermo_temp dynamic/dof yes",
                f"molecule mol mol.{self._mol}",
                f"group {self._mol} molecule 0",
                f"fix dpgcmc {self._mol} gcmc 1 {n_ex} {n_mc} 0 {seed()} "\
                        f"{T} 0.0 {dmax} mol mol pressure {P} full_energy",
                f"thermo {disp_freq}",
                f"run {equil_steps}",
                f"dump 1 all custom {write_freq} gcmc.dump id type x y z",
                f"run {steps}",
                ]

        return commands

    def process(self, file_out=None):
        atoms = read_dump("gcmc.dump", self.type_map)
        write(self.file_out, atoms)


Simulations = {
    "spe": SPE,
    "opt": Opt,
    "cellopt": CellOpt,
    "nvt-md": NVT,
    "npt-md": NPT,
    "eos": EOS,
    "vib": Vib,
    "gcmc": GCMC,
}



import numpy as np
from ase.calculators.calculator import Calculator, all_changes
from ase.units import GPa

from dptools.simulate.lammps_io import LammpsInput
from dptools.utils import read_type_map, graph2typemap

class LmpCalc(Calculator):
    """
    Base class for lammps ASE calculator.
    """

    name = "Lammps"
    implemented_properties = ["energy", "forces", "stress"]

    def __init__(self,
                 type_map=None,
                 minimize=False,
                 relax_cell=False,
                 ftol=1e-2,
                 lmp=None,
                 run_command=None,
                 verbose=False,
                 label="lamp",
                 **kwargs):

        Calculator.__init__(self, label=label, **kwargs)

        if lmp is None:
            import lammps
            if verbose:
                lmp = lammps.lammps()
            else:
                lmp = lammps.lammps(cmdargs="-screen none".split())

        if type_map is not None:
            self._type_map = read_type_map(type_map)
        if run_command is None and not (minimize or relax_cell):
            run_command = ["run 0"]
        elif isinstance(run_command, str):
            run_command = [run_command]
        elif relax_cell:
            run_command = ["fix cellopt all box/relax aniso 1.0",
                           "min_modify norm max",
                           f"minimize 0.0 {ftol} 1000 10000"]
        elif minimize:
            run_command = ["min_modify norm max",
                           f"minimize 0.0 {ftol} 1000 10000"]
        self.lmp = lmp
        self.minimize = minimize
        self.relax_cell = relax_cell
        self.ftol = ftol
        self.run_command = run_command

    def calculate(self,
                  atoms=None,
                  properties=["energy", "forces", "stress"],
                  system_changes=all_changes,
                  update=True,
                  ):
        """
        Run lammps calculation on.

        Args:
            atoms (ase.Atoms, optional): Atoms object to run the calculation on.
            properties (list[str]): unused, only for function signature compatibility,
            system_changes (list[str]): unused, only for function signature compatibility
            update (bool): Update atoms with new positions, energy, forces, etc. if True.
        """
        if atoms is not None:
            self.atoms = atoms.copy()

        # XXX: Is there a better way to do this? Always write input or no?
        self.write_input(atoms)

        for comm in self.run_command:
            self.lmp.command(comm)

        if update:
            # lammps is annoying and jumbles up indices sometimes
            ids = self.lmp.numpy.extract_atom("id").T[0]
            self._sort = np.argsort(ids)

            self.update_atoms(atoms)
            energy = self.lmp.get_thermo("etotal")
            forces = self.lmp.numpy.extract_atom("f")[self._sort]
            stress = self.read_stress()
            self.results["energy"] = energy
            self.results["forces"] = forces
            self.results["stress"] = stress

    def update_atoms(self, atoms):
        """
        Update positions of atoms and shift to origin if necessary after running calculation
        (i.e. align lammps positions with ASE unit cell).
        """
        positions = self.lmp.numpy.extract_atom("x")[self._sort]
        # shift positions to align with cell at origin
        shift = np.array([self.lmp.get_thermo(l) for l in ["xlo", "ylo", "zlo"]])
        atoms.positions = positions - shift
        atoms.cell = self.read_cell()

    def read_stress(self):
        """
        Returns:
            stress (np.array): Stress tensor in Voigt notation.
        """
        pxx = self.lmp.get_thermo("pxx") * 1e-3 # kBar
        pyy = self.lmp.get_thermo("pyy") * 1e-3
        pzz = self.lmp.get_thermo("pzz") * 1e-3
        pxy = self.lmp.get_thermo("pxy") * 1e-3
        pyz = self.lmp.get_thermo("pyz") * 1e-3
        pxz = self.lmp.get_thermo("pxz") * 1e-3
        stress = -np.array([pxx, pyy, pzz, pyz, pxz, pxy]) * 1e-1 * GPa
        return stress

    def read_cell(self):
        """
        Returns:
            (list): Cell lengths and angles to use with ASE
                [a, b, c, alpha, beta, gamma].
        """
        a = self.lmp.get_thermo("cella")
        b = self.lmp.get_thermo("cellb")
        c = self.lmp.get_thermo("cellc")
        alpha = self.lmp.get_thermo("cellalpha")
        beta = self.lmp.get_thermo("cellbeta")
        gamma = self.lmp.get_thermo("cellgamma")
        return [a, b, c, alpha, beta, gamma]

    def write_input(self):
        """System / forcefield specific"""
        pass


class DeepMD(LmpCalc):
    """
    ASE calculator to interface with LAMMPS and DeePMD-kit.

    Args:
        graph (str): Path to deepmd-kit .pb file with saved MLP model.

        type_map (dict): Dictionary of atom type mapping for graph. Should be able
            to infer automatically from graph, but for some reason (related to deepmd's tensorflow
            version, I think) lammps crashes if DeepPotential is imported before running.
            In short, type_map is not optional until that issue is resolved.

        minimize (bool): Run lammps geometry optimizaiton. Shortcut to be used instead of
            entering the minimize commands in run_command.

        relax_cell (bool): Run lammps unit cell optimization. Shortcut to be used instead of
            entering the box/relax fix and minimize commands in run_command.

        ftol (float): Force tolerance for lammps minimize convergence, only used if minimize
            or relax_cell are set to True.

        lmp (lammps.core.lammps): lammps instance (from lmp = lammps.lammps()).
            Creates new instance if None specified.

        run_command (list[str] or str): lammps commands to run simulation beyond the
            general setup commands obtained from lammps_io.LammpsInput

        verbose (bool): Display lammps output in console if True.
    """
    name = "DP"
    implemented_properties = ["energy", "forces", "stress"]
    def __init__(self,
                 graph,
                 type_map=None,
                 minimize=False,
                 relax_cell=False,
                 ftol=1e-2,
                 lmp=None,
                 run_command=None,
                 verbose=False,
                 label="lamp",
                 **kwargs):

        super().__init__(
                       type_map=type_map,
                       minimize=minimize,
                       relax_cell=relax_cell,
                       ftol=ftol,
                       lmp=lmp,
                       run_command=run_command,
                       verbose=verbose,
                       label=label,
                       **kwargs
                       )

        self.graph = graph
        self.style = f"deepmd {self.graph}"

    def set_atoms(self, atoms=None):
        self.set_types(atoms)
        for a in atoms:
            a.tag = self._type_map[a.symbol]

    def set_charges(self):
        self.charges = np.zeros(len(self.atoms))

    def set_types(self, atoms=None):
        if atoms is None:
            atoms = self.atoms

        if not hasattr(self, "_type_map"):
            raise NotImplementedError("Automatic detection of type_map not working, "
                        "please manually specify type_map")

            # TODO: Figure out why this fails, need to manually specify type_map until then
            # Automatically determines type_map from self.graph
            #from deepmd import DeepPotential
            #dp = DeepPotential(self.graph)
            #type_map = {sym: i for i, sym in enumerate(dp.get_type_map())}
            self._type_map = graph2typemap(self.graph)

        # lammps type indexing needs to start at 1, not 0
        if 0 in self._type_map.values():
            self._type_map = {sym: i + 1 for sym, i in self._type_map.items()}
        #for a in atoms:
        #    a.tag = self._type_map[a.symbol]

        # need to invert dict for lammps_io - a bit confusing, probably rework
        self.types = {v: k for k, v in self._type_map.items()}

    def set_coeffs(self):
        self.coeffs = None
        self.pair_coeffs = None

    def write_input(self, atoms=None):
        # XXX maybe add this to base class if identical for other FFs too
        if atoms is not None:
            self.atoms = atoms
        self.set_atoms(atoms)
        self.set_charges()
        self.set_coeffs()
        self.io = LammpsInput(self.atoms,
                              self.types,
                              charges=self.charges,
                              pair_style=self.style,
                              pair_coeff=self.pair_coeffs,
                              )
        self.io.write()
        self.lmp.command("clear")
        self.lmp.file("in.atoms")
        with open("in.atoms", "a") as file:
            for comm in self.run_command:
                file.write(f"{comm}\n")


DeePMD = DeepMD # I realized after a year that deepmd-kit capitalizes the P, oops


class ClayFF(LmpCalc):
    # Class for the classical ClayFF force field. Unused for deepmd but I will leave it here
    # for future reference
    name = "ClayFF"
    implemented_properties = ["energy", "forces", "stress"]
    def __init__(self,
                 type_map=None,
                 minimize=False,
                 relax_cell=False,
                 ftol=1e-2,
                 lmp=None,
                 run_command=None,
                 verbose=False,
                 label="lamp",
                 cutoff=12.0,
                 **kwargs):

        super().__init__(
                       type_map=type_map,
                       minimize=minimize,
                       relax_cell=relax_cell,
                       ftol=ftol,
                       lmp=lmp,
                       run_command=run_command,
                       verbose=verbose,
                       label=label,
                       **kwargs
                       )

        self.rc = cutoff
        self.style = f"lj/cut/coul/long {self.rc}"

    def set_atoms(self, atoms=None):
        if not hasattr(self, "_type_map"):
            self.set_types(atoms)
        if 0 in atoms.get_tags():
            for a in atoms:
                a.tag = self._type_map[a.symbol]
        #lengths = atoms.cell.lengths()
        #repeat = [int((self.rc * 2) // l + 1) for l in lengths]
        #if hasattr(self, 'repeat'):
        #    if repeat != self.repeat:
        #        raise Exception("REPEAT CHANGED, FIX IT OR FIX LOSS FCN")
        #self.repeat = repeat
        #atoms = atoms.repeat(self.repeat)
        #self.atoms = atoms

    def set_charges(self):
        q_vals = {"Si": 2.1, "O": -1.05}
        charges = [q_vals[a.symbol] for a in self.atoms]
        self.charges = np.array(charges)

    def set_types(self, atoms=None):
        if atoms is None:
            atoms = self.atoms
        symbols = np.unique(atoms.get_chemical_symbols())
        types = {s: i + 1 for i, s in enumerate(symbols)}
        for a in atoms:
            a.tag = types[a.symbol]
        self._type_map = types
        self.types = {v: k for k, v in types.items()} # inverting dict, probably stupid

    def set_coeffs(self):
        epsilon = {"Si": 7.981163385703463e-08, "O": 0.006738781799175867}
        sigma = {"Si": 3.302027, "O": 3.165541}
        self.coeffs = {k: {"eps": epsilon[v], "sig": sigma[v]}
                        for k, v in self.types.items()}

    def write_input(self, atoms=None):
        if atoms is not None:
            self.atoms = atoms
        self._check_cell()
        self.set_atoms(atoms)
        self.set_charges()
        self.set_coeffs()
        self.io = LammpsInput(self.atoms,
                              self.types,
                              charges=self.charges,
                              pair_style=self.style,
                              coeffs=self.coeffs,
                              cutoff=self.rc)

        self.io.write_atoms(self.atoms)
        self.io.write_input()
        self.lmp.command("clear")
        self.lmp.file("in.atoms")
        if self.relax_cell:
            self.lmp.command("fix 1 all box/relax tri 1.0")

    def _check_cell(self):
        """Checks if lattice constants are > 2 * cutoff"""
        min_lc = min(self.atoms.cell.lengths())
        if min_lc < 2 * self.rc:
            raise Exception("Lattice constant < 2 * cutoff\nRepeat Atoms or decrease cutoff")


class BKS(LmpCalc):
    # Class for the classical BKS force field. Unused for deepmd but I will leave it here
    # for future reference
    name = "BKS"
    implemented_properties = ["energy", "forces", "stress"]
    def __init__(self,
                 type_map=None,
                 minimize=False,
                 relax_cell=False,
                 ftol=1e-2,
                 lmp=None,
                 run_command=None,
                 verbose=False,
                 label="lamp",
                 cutoff=12.0,
                 **kwargs):

        super().__init__(
                       type_map=type_map,
                       minimize=minimize,
                       relax_cell=relax_cell,
                       ftol=ftol,
                       lmp=lmp,
                       run_command=run_command,
                       verbose=verbose,
                       label=label,
                       **kwargs
                       )

        self.rc = cutoff
        self.style = f"buck/coul/long {self.rc}"

    def set_atoms(self, atoms=None):
        if not hasattr(self, "_type_map"):
            self.set_types(atoms)
        if 0 in atoms.get_tags():
            for a in atoms:
                a.tag = self._type_map[a.symbol]
        #lengths = atoms.cell.lengths()
        #repeat = [int((self.rc * 2) // l + 1) for l in lengths]
        #if hasattr(self, 'repeat'):
        #    if repeat != self.repeat:
        #        raise Exception("REPEAT CHANGED, FIX IT OR FIX LOSS FCN")
        #self.repeat = repeat
        #atoms = atoms.repeat(self.repeat)
        #self.atoms = atoms

    def set_charges(self):
        q_vals = {"Si": 2.4, "O": -1.2}
        charges = [q_vals[a.symbol] for a in self.atoms]
        self.charges = np.array(charges)

    def set_types(self, atoms=None):
        if atoms is None:
            atoms = self.atoms
        symbols = np.unique(atoms.get_chemical_symbols())
        types = {s: i + 1 for i, s in enumerate(symbols)}
        for a in atoms:
            a.tag = types[a.symbol]
        self._type_map = types
        self.types = {v: k for k, v in types.items()} # inverting dict, probably stupid

    def set_coeffs(self):
        self.coeffs = None

        Aij = {"Si-O": 18003.7572, "O-O": 1388.7730, "Si-Si": 0.0}
        bij = {"Si-O": 1 / 4.87318, "O-O": 1 / 2.760, "Si-Si": 1.0}
        cij = {"Si-O": 133.5381, "O-O": 175.000, "Si-Si": 0.0}

        pair_coeffs = ""
        for k in Aij:
            ij = [self._type_map[k.split("-")[0]], self._type_map[k.split("-")[1]]]
            if ij[0] > ij[1]:
                ij.reverse()
            pair_coeffs += f"pair_coeff {ij[0]} {ij[1]} {Aij[k]} {bij[k]} {cij[k]}\n"

        self.pair_coeffs = pair_coeffs

    def write_input(self, atoms=None):
        # XXX maybe add this to base class if identical for other FFs too
        if atoms is not None:
            self.atoms = atoms
        self._check_cell()
        self.set_atoms(atoms)
        self.set_charges()
        self.set_coeffs()
        self.io = LammpsInput(self.atoms,
                              self.types,
                              charges=self.charges,
                              pair_style=self.style,
                              coeffs=self.coeffs,
                              pair_coeffs=self.pair_coeffs,
                              cutoff=self.rc)

        self.io.write_atoms(self.atoms)
        self.io.write_input()
        self.lmp.command("clear")
        self.lmp.file("in.atoms")
        if self.relax_cell:
            self.lmp.command("fix 1 all box/relax tri 1.0")

    def _check_cell(self):
        """Checks if lattice constants are > 2 * cutoff"""
        min_lc = min(self.atoms.cell.lengths())
        if min_lc < 2 * self.rc:
            raise Exception("Lattice constant < 2 * cutoff\nRepeat Atoms or decrease cutoff")



import os
import glob
import json

from dptools.cli import BaseCLI
from dptools.utils import randomize_seed
from dptools.hpc import SlurmJob


class CLI(BaseCLI):
    """
    Setup and submit jobs to train single or ensemble of deepmd-kit models.

    :doc:`Complete documentation here<../commands/train>`

    Examples:

    .. code-block:: console

        $ dptools train /path/to/dataset # simple single model
        $ dptools train -e /path/to/dataset # ensemble (4) of models
        $ dptools train -e -s /path/to/dataset # submit 4 slurm jobs to train ensemble
        $ dptools train -p /path/to/training/dir /path/to/dataset # specify dir to train in
        $ dptools train -i /path/to/in.json /path/to/dataset # specify in.json parameter file
    """

    help_info = "Setup and submit jobs to train deepmd-kit models"

    def add_args(self):
        self.parser.add_argument("dataset", type=str,
                help="Path to dataset parent directory")
        self.parser.add_argument("-e", "--ensemble", action="store_true",
                help="Make ensemble (4) of DP models to train")
        self.parser.add_argument("-s", "--submit", action="store_true",
                help="Automatically submit slurm job(s) to train model(s)")
        self.parser.add_argument("-p", "--path", type=str, default=".",
                help="Specify path to training directory")
        self.parser.add_argument("-i", "--input", type=str, default=None,
                help="Specify path to in.json deepmd parameter file to use for training")

    def main(self, args):
        if args.dataset == ".":
            raise ValueError("Do not train inside the dataset folder"\
                    " (dataset can not be '.')")
        self.datapath = os.path.abspath(args.dataset)
        self.path = os.path.abspath(args.path)
        if args.input:
            self._json = os.path.abspath(args.input)
        else:
            default = os.path.dirname(os.path.abspath(__file__))
            self._json = os.path.join(default, "../train/in.json")

        if args.ensemble:
            ens_dirs = ["00", "01", "02", "03"]
            self.dirs = [os.path.join(self.path, d) for d in ens_dirs]
        else:
            self.dirs = [self.path]
        self._sub = args.submit
        self.setup()
        self.submit_jobs()

    def setup(self):
        """
        Loads training parameter json file, updates with relevant info
        (random seeds, training/validation dirs, type map), and writes to
        training dir.
        """

        with open(self._json, "r") as file:
            in_json = json.loads(file.read())
        for d in self.dirs:
            jsn = randomize_seed(in_json)
            jsn = self.link_dirs(jsn)
            jsn = self.set_types(jsn)
            self.write_json(jsn, d)

    @staticmethod
    def write_json(src, dest):
        """
        Writes training parameter json file to specified training directory.

        Args:
            src (dict): jsonnable dictionary with deepmd-kit training parameters.
            dest (str): Path to training directory to write .json file to
        """
        os.makedirs(dest, exist_ok=True)
        file_path = os.path.join(dest, "in.json")
        with open(file_path, "w") as file:
            file.write(json.dumps(src, indent=4))

    def link_dirs(self, in_json):
        """
        Find and add all system training and validation set folders found in dataset
        directory to in_json.

        Args:
            in_json (dict): jsonnable dictionary with deepmd-kit training parameters.
        """
        possible_dirs = sorted(glob.glob(f"{self.datapath}/*"))
        dirs = [d for d in possible_dirs if self._check_dir(d)]
        train = [os.path.join(d, "train") for d in dirs]
        validation = [os.path.join(d, "validation") for d in dirs]
        in_json["training"]["training_data"]["systems"] = train
        in_json["training"]["validation_data"]["systems"] = validation
        return in_json

    def set_types(self, in_json):
        """
        Load type_map.json from dataset directory and add types to in_json.

        Args:
            in_json (dict): jsonnable dictionary with deepmd-kit training parameters.
        """
        if not hasattr(self, "types"): # only need to read type_map once
            type_map_path = os.path.join(self.datapath, "type_map.json")
            with open(type_map_path, "r") as file:
                type_map = json.loads(file.read())
            self.types = [type_map[str(i)] for i in range(len(type_map))]

        in_json["model"]["type_map"] = self.types
        return in_json

    @staticmethod
    def _check_dir(directory):
        """
        Checks if directory is part of the training dataset or not.

        Args:
            directory (str): Path to directory to check.

        Returns:
            bool: True if directory contains training/validation data, else False.
        """
        if not os.path.isdir(directory):
            return False
        check1 = "train" in os.listdir(directory)
        check2 = "validation" in os.listdir(directory)
        return check1 and check2

    def submit_jobs(self):
        """
        Write and submit Slurm job(s) to train model or ensemble of models.
        """
        from dptools.env import get_dpfaults
        hpc_info = get_dpfaults(key="sbatch")
        sbatch_comment = hpc_info.pop("SBATCH_COMMENT")
        commands = ["dp train in.json", "dp freeze -o graph.pb"]
        jobs = SlurmJob(sbatch_comment,
                        commands=commands,
                        directories=self.dirs,
                        file_name="dptools.train.sh",
                        **hpc_info)
        jobs.write(sub=self._sub)



import os
import numpy as np
from ase.io import write

from dptools.train.ensemble import SampleConfigs
from dptools.cli import BaseCLI
from dptools.utils import graph2typemap, read_type_map
from dptools.env import load, get_dpfaults

class CLI(BaseCLI):
    """
    Intelligently select new training configurations using an ensemble of models
    and the approach described in DP-GEN (DOI: 10.1016/j.cpc.2020.107206).

    :doc:`Complete documentation here<../commands/sample>`

    Examples:

    .. code-block:: console

        $ dptools sample -n 200 nvt-md.traj
        $ dptools sample -n 100 --lo 0.05 --hi 0.25 nvt-md.traj
        $ dptools sample -m water_ensemble -p npt-md.traj

    """
    help_info = "Select new training configs from MD traj "\
        "using force prediction deviations from ensemble of DPs"

    def add_args(self):
        self.parser.add_argument(
            "configurations",
            nargs="+",
            help="Snapshots from MD simulation to select new training configuraitons "\
                "from (.traj or similar)"
        )
        self.parser.add_argument("-m", "--model-ensemble", nargs="+", type=str,
                help="Paths to ensemble of models or label of set models")
        self.parser.add_argument("-n", type=int, default=300,
                help="Max number of new configurations to select")
        self.parser.add_argument("--lo", type=float, default=0.05,
                help="Min value of eps_t (force dev) to select new configs from")
        self.parser.add_argument("--hi", type=float, default=0.35,
                help="Max value of eps_t (force dev) to select new configs from")
        self.parser.add_argument("-o", "--output", nargs=1, type=str, default="new_configs.traj",
                help="File to write new configurations to")
        self.parser.add_argument("-p", "--plot-dev", action="store_true",
                help="Plot histogram of max force deviation of model ensemble for each config")
        self.parser.add_argument("--plot-steps", action="store_true",
                help="Plot dev versus number of steps")


    def main(self, args):
        self.outfile = os.path.basename(args.output)
        self.load_ensemble(args.model_ensemble) # sets self.type_map and self.graphs
        self.set_configs(args.configurations)
        self.devs = [] # max force deviation of model ensemble

        wd = os.getcwd()
        for configs, dir in zip(self.configs, self.dirs):
            os.chdir(dir)

            self.sample(configs, args)

            os.chdir(wd)

        if args.plot_dev or args.plot_steps:
            self.plot(steps=args.plot_steps)

    def load_ensemble(self, ensemble):
        if not ensemble or len(ensemble) == 1:
            if ensemble is not None and len(ensemble) == 1:
                load(ensemble[0])
            defaults = get_dpfaults(key="ensemble")
            self.type_map, *ensemble = defaults
        else:
            self.type_map = graph2typemap(ensemble[0])
        self.ensemble = ensemble

    def set_configs(self, configs):
        self.configs = [os.path.abspath(c) for c in configs]
        if len(configs) == 1:
            dirs = ["."]
        else:
            dirs = [os.path.dirname(c) for c in self.configs]
            if len(np.unique(dirs)) != len(self.configs):
                # FIXME: Results are overwritten if multiple structure inputs are in the same dir
                raise Exception("Can't resolve inputs, harass me to fix this")
        self.dirs = dirs

    def sample(self, configs, args):
        self.sampler = SampleConfigs(configs, self.ensemble, read_type_map(self.type_map))
        new_configs = self.sampler.sample(lo=args.lo, hi=args.hi, n=args.n)

        self.devs.append(self.sampler.dev)
        write(self.outfile, new_configs)

    def plot(self, steps=False):
        import matplotlib.pyplot as plt
        fig, ax = plt.subplots(figsize=(5.5, 4))
        for dev, _dir in zip(self.devs, self.dirs):
            ax = self.sampler.plot(dev=dev, steps=steps, ax=ax, label=os.path.relpath(_dir))
        if len(self.dirs) > 1:
            ax.legend()
        plt.show()



"""
Functions for interacting with simulation parameter files (.yaml).
If adding a new simulation parameter, add a description/usage hint to
dptools.parameters.descriptions, and the hint will appear in params.yaml.
"""
import os
import requests
from ruamel.yaml import YAML

basedir = os.path.abspath(os.path.dirname(__file__))
param_file = os.path.join(basedir, "parameter_sets.yaml")

descriptions = {
    "type": "Type of calculation (spe, opt, cellopt, nvt-md, npt-md, eos)",
    "disp_freq": "Print lammps output every {disp_freq} steps (thermo disp_freq)",
    "nsw": "Max number of iterations",
    "ftol": "Force convergence tolerance for lammps minimize",
    "etol": "Energy convergence tolerance for lammps minimize",
    "opt_type": "(iso, aniso, tri) see https://docs.lammps.org/fix_box_relax.html",
    "Ptarget": "Desired pressure [bar]",
    "steps": "Total number of timesteps to run simulation",
    "timestep": "[fs]",
    "equil_steps": "Number of timesteps to run initial equilibration at Ti",
    "Ti": "Initial temperature [K] at start of simulation",
    "Tf": "Final temperature [K] of simulation (ramped from Ti to Tf)",
    "Pi": "Initial pressure [bar] at start of simulation",
    "Pf": "Final pressure [bar] of simulation (ramped from Pi to Pf)",
    "pre_opt": "Optimize structure (and cell for npt-md) before starting MD run",
    "write_freq": "Write MD image every {write_freq} steps",
    "N": "Create N equally spaced structures with cell Volumes from V0*lo to V0*hi",
    "lo": "Lower bound for cell deformations (min volume = lo * V0)",
    "hi": "Upper bound for cell deformations (max volume = hi * V0)",
    "nfree": "(2 or 4) Number of displacements for each degree of freedom",
    "delta": "Magnitude of each displacement [Å]",
}


def write_yaml(param_dict, file):
    """
    Write simulation parameter set to yaml file. Used by the :doc:`../commands/get` to
    retrieve params.yaml files for specific simulation to be edited and then either used
    with :doc:`../commands/set` or :doc:`../commands/run` command.

    Args:
        param_dict (dict): Dictionary containing each parameter and its
            corresponding value.
        file (file object): File to write simulation parameter set to
            (generally params.yaml)
    """
    lengths = [len(k) + len(str(v)) for k, v in param_dict.items()]
    column = max(lengths) + 4  # align parameter description comments
    for param in param_dict:
        description = descriptions.get(param, f"No description available for parameter '{param}'")
        param_dict.yaml_add_eol_comment(description, key=param, column=column)
    YAML().dump(param_dict, file)


def get_parameter_sets():
    """
    Load simulation parameter sets from parameter_sets.yaml.
    """
    with open(param_file) as file:
        parameter_sets = YAML().load(file.read())
    return parameter_sets


def set_parameter_set(param_dict):
    """
    Write simulation parameter set to parameter_sets.yaml for future use.

    Args:
        param_dict (dict): Dictionary containing each parameter and its
            corresponding value.
    """
    if isinstance(param_dict, str):
        with open(param_dict) as file:
            param_dict = YAML(typ="safe").load(file.read())
    parameter_sets = get_parameter_sets()
    calc_type = param_dict.get("type")
    parameter_sets[calc_type] = param_dict
    with open(param_file, "w") as file:
        YAML().dump(parameter_sets, file)


def reset_params():
    url = "https://github.com/tysours/DPTools/raw/main/dptools/simulate/parameter_sets.yaml"
    req = requests.get(url, allow_redirects=True)
    parameter_sets = YAML().load(req.content)
    with open(param_file, "w") as file:
        YAML().dump(parameter_sets, file)



import os

from dptools.cli import BaseCLI
from dptools.env import load, set_model, set_sbatch, set_params, set_training_params


class CLI(BaseCLI):
    """
    Set DP models, Slurm settings, calculation parameters, or deepmd-kit training parameters

    :doc:`Complete documentation here<../commands/set>`

    Examples:

    .. code-block:: console

        $ dptools set /home/user/projects/dp_models/h2o.pb
        $ dptools set 00/graph.pb 01/graph.pb 02/graph.pb 03/graph.pb # ensemble
        $ dptools set -m water /home/user/projects/dp_models/h2o.pb
        $ dptools set -m hpc_para parallel_settings.sh
        $ dptools set -m hpc_serial serial_settings.sh
        $ dptools set params.yaml
        $ dptools set /path/to/in.json
    """
    help_info = "Set DP model defaults, calculation parameters, or sbatch settings"

    def add_args(self):
        _help = "Path to DP model, params.yaml, or {script}.sh to set as default.\n"\
             "Need .pb, .yaml, or .sh extension to set model, params, or sbatch, respectively."
        self.parser.add_argument(
            "thing",
            nargs="+",
            help=_help,
        )
        self.parser.add_argument("-m", "--model-label", type=str, default=None,
                help="Save model/sbatch with specific label to access during "\
                        "'dptools run -m {label} {calc} {structure}'")

    def main(self, args):
        if args.model_label:
            load(args.model_label)
        for i, t in enumerate(args.thing):
            if t.endswith(".pb") and i > 0:
                self.set(t, n_model=i+1)
            else:
                self.set(t)

    def set(self, thing, **kwargs):
        """
        Set file parameters to .env or simulation parameters to yaml file.

        Note:
            Multiple .pb inputs allowed for setting model ensembles.

        Args:
            thing (str): Path to file you want to set (.pb, .sh, .yaml, or .json file).
        """
        ext2function = {
                        ".pb": set_model,
                        ".sh": set_sbatch,
                        ".yaml": set_params,
                        ".json": set_training_params,
                        }

        ext = os.path.splitext(thing)[-1]
        if ext not in ext2function:
            raise TypeError(f"Unrecognized file type for {thing}. Try 'dptools set -h'")
        set_thing = ext2function[ext]
        set_thing(thing, **kwargs)



import os
import shutil

from dptools.cli import BaseCLI
from dptools.simulate.parameters import get_parameter_sets, write_yaml

class CLI(BaseCLI):
    """
    Get params.yaml for specific simulation type. Can also be used to get
    in.json file (deepmd-kit training parameters).

    :doc:`Complete documentation here<../commands/get>`

    Examples:

    .. code-block:: console

        $ dptools get cellopt
        $ dptools get list # see available simulation options
        $ dptools get nvt-md
        $ dptools get nvt-md.900K # custom simulation params
        $ dptools get in.json # get training param file
    """
    help_info = "Get params.yaml for specific simulation type"
    def add_args(self):
        self.parser.add_argument(
            "simulation",
            type=str,
            help="Simulation type to generate params.yaml file "
            "(spe, opt, cellopt, nvt-md, npt-md, eos, vib)."
            "\nCan also specify label of saved simulations (e.g. nvt-md.label)",
        )

    def main(self, args):
        if args.simulation.endswith(".json"):
            here = os.path.dirname(__file__)
            json_path = os.path.abspath(os.path.join(here, "../train/in.json"))
            shutil.copy(json_path, ".")
            return

        param_sets = get_parameter_sets()
        if args.simulation == "list":
            print()
            print("# Available simulation types:")
            for key in param_sets.keys():
                print(key)
            print()
            return

        params = param_sets[args.simulation]
        with open("params.yaml", "w") as file:
            write_yaml(params, file)



"""
Assorted utilities for plots, reading things, converting things, etc.
"""
import numpy as np
from ase import Atoms
from ase.io import write, read
from ase.io.formats import string2index
from ase.db import connect
from ase.data import chemical_symbols
import json

#seaborn.color_palette('deep')
colors = [(0.2980392156862745, 0.4470588235294118, 0.6901960784313725),
          (0.8666666666666667, 0.5176470588235295, 0.3215686274509804),
          (0.3333333333333333, 0.6588235294117647, 0.40784313725490196),
          (0.7686274509803922, 0.3058823529411765, 0.3215686274509804),
          (0.5058823529411764, 0.4470588235294118, 0.7019607843137254),
          (0.5764705882352941, 0.47058823529411764, 0.3764705882352941),
          (0.8549019607843137, 0.5450980392156862, 0.7647058823529411),
          (0.5490196078431373, 0.5490196078431373, 0.5490196078431373),
          (0.8, 0.7254901960784313, 0.4549019607843137),
          (0.39215686274509803, 0.7098039215686275, 0.803921568627451)]

def _generate_color():
    for color in colors:
        yield color


_gen_color = _generate_color()
def next_color():
    """
    Convenience function to grab a new color during plotting loops.

    Returns:
        color (tuple[float]): New color rgb vals from seaborn.color_palette('deep')

    Example:

        .. code-block:: python

            for values in all_values:
                plt.plot(values, color=next_color())
    """
    return next(_gen_color)


def read_dump(dump, type_map, index=":"):
    """
    Reads in lammps dump file and returns corresponding ase.Atoms list.

    Args:
        dump (str): Path to dump file to read.
        type_map (dict): Dictionary with element-index mapping, e.g. {'Si': 0, 'O': 1}
        index (str): index slice to control which images are returned, e.g. ':', '::100', etc.

    Returns:
        traj (list[ase.Atoms]): List of dump images as ase.Atoms objects
    """
    type_map = read_type_map(type_map) # support str, json, dict inputs
    if 0 in type_map.values(): # lammps indexing starts at 1
        type_map = {k: v + 1 for k, v in type_map.items()}
    type_map = {v: k for k, v in type_map.items()} # invert to find symbol from type index
    with open(dump) as file:
        lines = file.readlines()
    traj = []
    # n_atoms = int(lines[3])
    for i, line in enumerate(lines):
        if "BOX BOUNDS" in line:
            n_atoms = int(lines[i - 1])
            lammps_cell = np.array([_str_to_float(lines[i + j + 1]) for j in range(3)])
            cell, shift = convert_dump_cell(lammps_cell)
        elif "ITEM: ATOMS" in line:
            ids = [int(l.split()[0]) for l in lines[i + 1 : i + 1 + n_atoms]]
            sort = np.argsort(ids)
            # TODO: replace index hardcodinng
            types = np.array(
                [int(l.split()[-4]) for l in lines[i + 1 : i + 1 + n_atoms]]
            )
            symbols = np.array([type_map[t] for t in types])
            positions = np.array(
                [_str_to_float(lines[i + j + 1]) for j in range(n_atoms)]
            )
            if "xs" in line:
                positions = positions @ cell
            positions = positions - shift  # shift atoms to origin for ASE Atoms object
            atoms = Atoms(
                positions=positions[sort], symbols=symbols[sort], cell=cell, pbc=True
            )
            atoms.set_tags(types[sort])
            traj.append(atoms)
    return traj[string2index(index)]


def _str_to_float(l):
    return list(map(float, l.split()[-3:]))


def convert_dump_cell(lammps_cell):
    """
    Converts lammps dump cell format to ase cell.

    Args:
        lammps_cell (numpy.array): Simulation cell from lammps dump file.

    Returns:
        cell (numpy.array): New 3x3 array to use with ase.
        shift (numpy.array): xyz shifts to place atoms at origin aligned with cell.
    """
    xlo_bound, xhi_bound = lammps_cell[0, :2]
    ylo_bound, yhi_bound = lammps_cell[1, :2]
    zlo_bound, zhi_bound = lammps_cell[2, :2]
    if lammps_cell.shape == (3, 3):
        xy, xz, yz = lammps_cell[:, -1]
    elif lammps_cell.shape == (3, 2):
        xy, xz, yz = 0.0, 0.0, 0.0

    xlo = xlo_bound - min((0.0, xy, xz, xy + xz))
    xhi = xhi_bound - max((0.0, xy, xz, xy + xz))
    ylo = ylo_bound - min((0.0, yz))
    yhi = yhi_bound - max((0.0, yz))
    zlo = zlo_bound
    zhi = zhi_bound

    a = [xhi - xlo, 0, 0]
    b = [xy, yhi - ylo, 0]
    c = [xz, yz, zhi - zlo]
    cell = np.array([a, b, c])
    shift = np.array([xlo, ylo, zlo])
    return cell, shift


def tag_groups(atoms, mult=1.0, retag=True):
    """
    Takes Atoms object of disconnected groups (e.g. adsorbates in MOF) and assigns a unique tag
    to each group. Returns list of nested lists with indices corresponding to each tag

    Args:
        atoms (ase.Atoms): Atoms object to tag.
        mult (float): Coefficient for adjusting neighbor search distances
            (ase.neighborlist.natural_cutoffs function).
        retag (bool): If false, a copy of atoms is used for tagging, leaving the original
            tags unmodified.

    Returns:
        groups (dict): dict of all indices and elements corresponding to
            each group of tagged atoms.
    """
    from ase.neighborlist import NeighborList
    from ase.neighborlist import natural_cutoffs
    if not retag:
        atoms = atoms.copy()
    nl = NeighborList(natural_cutoffs(atoms, mult=mult), self_interaction=False, bothways=True)
    nl.update(atoms)
    atoms.set_tags(0) #initialize all tags to 0
    tag_index = 1
    groups = {}
    # Grabs atom from subset, tags all connected atoms
    # with tag_index until all atoms in subset are tagged. Repeats until no atoms are untagged
    while 0 in atoms.get_tags():
        groups[tag_index] = {}
        possible_atoms = [a.index for a in atoms if a.tag == 0] # all current untagged atoms
        neighbor_indices = [possible_atoms[0]] # grabs first untagged atom
        for i in neighbor_indices: # finds all connected atoms, tags, and adds to list of iterables
            new_atoms = [k for k in list(nl.get_neighbors(i)[0]) if atoms[k].tag == 0]
            atoms[i].tag = tag_index
            for j in new_atoms:
                atoms[j].tag = tag_index
            neighbor_indices.extend(new_atoms)

        groups[tag_index]["indices"] = [a.index for a in atoms if a.tag == tag_index]
        symbols = [a.symbol for a in atoms if a.tag == tag_index]
        groups[tag_index]["elements"] = list(np.unique(symbols))
        tag_index += 1

    return groups


def read_db(db_name, indices):
    """
    Reads ase db and returns entries as list of Atoms objects.
    Useless, just use ase.io.read(db_name, index=indices) instead.

    Args:
        db_name (str): Path to ase db file.
        indices (str): Index slice, e.g. ::10, :, 1:100, etc.

    Returns:
        traj (list[ase.Atoms]): db entries as list of Atoms objects.
    """
    with connect(db_name) as db:
        traj = [row.toatoms() for row in db.select()]
    return traj[string2index(indices)]


def graph2typemap(graph):
    """
    Determine type_map for a given deepmd model.

    Args:
        graph (str): Path to deepmd model .pb file.

    Returns:
        type_map (dict): Dictionary that maps each atomic symbol to type map index
    """
    from deepmd import DeepPotential
    dp = DeepPotential(graph)
    type_map = {sym: i for i, sym in enumerate(dp.get_type_map())}
    return type_map


def read_type_map(type_map_json):
    if isinstance(type_map_json, dict):
        type_map = type_map_json
    elif isinstance(type_map_json, str):
        if type_map_json.endswith(".json"):
            with open(type_map_json) as file:
                type_map = json.loads(file.read())
        else:
            type_map = str2typemap(type_map_json)
    else:
        if type_map_json is None:
            msg = "No default model set, try using:\tdptools set /path/to/graph.pb"
        else:
            msg = f"Unknown type_map format provided: {type_map_json}"
        raise TypeError(msg)
    return check_type_map(type_map)


def check_type_map(type_map_dict):
    # I'm inconsistent with key-value/value-key when writing type_map
    # so check if need to invert dict such that chem symbols are keys
    # TODO: fix inconsistencies
    invert = False
    for k, v in type_map_dict.items():
        if k not in chemical_symbols:
            invert = True
        break
    if invert:
        type_map_dict = {v: k for k, v in type_map_dict.items()}
    return type_map_dict

def typemap2str(type_map_dict):
    type_map = check_type_map(type_map_dict)
    tm_str = ",".join([f"{k}:{v}" for k, v in type_map.items()])
    return tm_str


def str2typemap(tm_str):
    items = tm_str.split(",")
    type_map = {i.split(":")[0]: int(i.split(":")[-1]) for i in items}
    return type_map


def randomize_seed(in_json):
    """
    Take input .json or dict with training parameters and randomize seeds
    for model ensemble training.

    Args:
        in_json (str or dict): Path to .json or dict with deepmd training parameters.

    Returns:
        (dict): Training parameter dictionary with new randomized seed values.
    """
    if isinstance(in_json, str):
        with open(in_json) as file:
            in_json = json.loads(file.read())
    elif not isinstance(in_json, dict):
        raise TypeError("Need dict or .json file to randomize seeds")
    seeds = get_seed(n=3)
    in_json["model"]["descriptor"]["seed"] = seeds[0]
    in_json["model"]["fitting_net"]["seed"] = seeds[1]
    in_json["training"]["seed"] = seeds[2]
    return in_json


def get_seed(max_val=999999, n=1):
    """
    Returns:
        seed (int or list[int]): n random ints between 0 and max_val.
    """
    if n > 1:
        seed = np.random.randint(max_val, size=n)
        seed = [int(s) for s in seed] # dtype=int64 not json serializable
    elif n == 1:
        seed = np.random.randint(max_val) # avoids single item array
    return seed


def columnize(*data):
    """
    Takes lists or 1D arrays and concatenates everything into columnized array.
    Basically just np.column_stack without needing a single tuple arg (i.e. slightly useless).
    """
    return np.array(list(data)).T


class Converter:
    """
    Class to convert between different ASE/VASP/LAMMPS outputs. Mostly useful if wanting
    to concatenate all MD images from variable-time jobs, but also supports lammps dump
    conversions and common ASE formats.

    Args:
        inputs (list[str]): Paths to input structure files to convert. Images from each
            input are concatenated in output.

        output (str): Name of file with desired conversion extension specified,
            e.g. ``'out.traj'``

        indices (str): Index slice, e.g. ``'::10'``, :, ``'1:100'``, etc.

    Example:

        .. code-block:: python

            # concat MD images (ignores 1st image if identical to final image in previous input)
            >>> from dptools.utils import Converter
            >>> inputs = ['md_000/vasprun.xml', 'md_001/vasprun.xml', 'md_002/vasprun.xml']
            >>> output = 'md.traj'
            >>> converter = Converter(inputs, outputs)
            >>> converter.convert()
    """

    def __init__(self, inputs, output, indices=":"):
        self.inputs = inputs
        self.output = output
        self.indices = indices
        self.type_readers = {
                             "xml": read,
                             "traj": read,
                             "db": read,
                             "cif": read,
                             "xyz": read,
                             "dump": read_dump,
                             "OUTCAR": read,
                             "POSCAR": read,
                             "CONTCAR": read,
                             "vasp": read,
                             }
        self.set_types()
        self.set_reader()

    def set_types(self):
        self.types = {"output": self._check_type(self.output),
                      "inputs": [self._check_type(i) for i in self.inputs]}
        if len(np.unique(self.types["inputs"])) > 1:
            raise ValueError(
                 f"multiple input types detected:\t{self.types['inputs']}\n"
            )

    def set_reader(self):
        self.reader = self.type_readers[self.types["inputs"][0]]

    def _check_type(self, f):
        ftype = f.split(".")[-1]
        if ftype not in self.type_readers:
            types = self.type_readers.items()
            raise NotImplementedError(f"supported types:\t{types}\nharass me for others")
        return ftype

    def read(self, **kwargs):
        self.atoms = []
        for i in self.inputs:
            atoms = self.reader(i, index=self.indices, **kwargs)
            if len(self.atoms) > 0 and len(atoms) > 1:
                # check to see if first image is identical to last image of previous file
                # primarily for concatenating MD runs from flex/overrun jobs
                pos1 = self.atoms[-1].positions
                pos2 = atoms[0].positions
                if (pos1 == pos2).all():
                    atoms = atoms[1:]
            self.atoms.extend(atoms)

    def convert(self, **kwargs):
        if not hasattr(self, "atoms"):
            self.read(**kwargs)
        write(self.output, self.atoms)



import os
import json

from dptools.cli import BaseCLI
from dptools.train.parity import EvaluateDP

class CLI(BaseCLI):
    """
    Generate parity plots comparing DP model accuracy for energy and force
    (and stress if available) predictions with ab-initio values.

    :doc:`Complete documentation here<../commands/parity>`

    Examples:

    .. code-block:: console

        $ dptools parity
        $ dptools parity /path/to/dataset/system*/test/set*
        $ dptools parity -m ../old_graph.pb test_set.traj
        $ dptools parity -l mae
    """
    help_info = "Generate energy and force (and stress if available) prediction parity plots for DP model"
    def add_args(self):
        # TODO: add more optional args (e.g. save plot)
        self.parser.add_argument("systems", nargs="*", metavar="system", help="Paths to deepmd-kit dataset folders, .traj, .db, etc.")
        self.parser.add_argument("-m", "--model", type=str, default="./graph.pb",
                help="Specify path of frozen .pb deepmd model to use")
        self.parser.add_argument("-l", "--loss-function", type=str, default="mse", choices=["mse", "mae", "rmse"],
                help="Type of loss function to display for parity plot error")
        self.parser.add_argument("-s", "--save-plot", type=str, metavar="file_name",
                help="Name of file (with extension) to save parity plot to.")
        self.parser.add_argument("-r", "--rasterized", action="store_true",
                help="Rasterize plot data to reduce size of file")
        self.parser.add_argument("--per-atom", action="store_true",
                help="Normalize energies per number of atoms")
        self.parser.add_argument("--xyz", action="store_true",
                help="Plot each xyz force component separately")
        self.parser.add_argument("--fancy", action="store_true",
                help="Create fancy density heat map for forces parity plot")

    def main(self, args):
        if len(args.systems) > 0:
            systems = args.systems
        else:
            systems = self.read_systems()
        evaldp = EvaluateDP(systems, dp_graph=args.model, per_atom=args.per_atom)
        evaldp.plot(
                loss=args.loss_function,
                xyz=args.xyz,
                fancy=args.fancy,
                save_file=args.save_plot,
                rasterized=args.rasterized,
            )

    @staticmethod
    def read_systems():
        """
        Load test sets from system paths in in.json training parameter file.

        Returns:
            list of str: Paths to system test sets found in training json file.
        """

        if "out.json" not in os.listdir() and "in.json" not in os.listdir():
            raise FileNotFoundError("Systems not specified and no in.json in $PWD")
        in_file = "in.json" if "in.json" in os.listdir() else "out.json"
        with open("in.json") as file:
            params = json.loads(file.read())
        systems = params["training"]["training_data"]["systems"]
        return [f"{s.split('/train')[0]}/test/set.000" for s in systems]



"""
Module for controlling HPC job submissions. Currently only supports Slurm jobs.
"""
import os

# defaults for Kulkarni group hpc systems
hpc_defaults = {
        "cori": {
            "SBATCH_COMMENT": "#SBATCH -J job -q shared -N 1 -t 11:00:00 "\
                "-C haswell --output=job.out --error=job.err --ntasks=1 -c 1",
            "OMP_NUM_THREADS": "1",
            "TF_INTRA_OP_PARALLELISM_THREADS": "1",
            "TF_INTER_OP_PARALLELISM_THREADS": "1"
            },
        "hpc": {
            "SBATCH_COMMENT": "#SBATCH --partition=med -N 1 --ntasks-per-node=8 "\
                "--output=job.out --error=job.err -t 96:00:00",
            "OMP_NUM_THREADS": "8",
            "TF_INTRA_OP_PARALLELISM_THREADS": "6",
            "TF_INTER_OP_PARALLELISM_THREADS": "2"
            },
        }

# TODO: Split sbatch into n_nodes, n_tasks_per_node, etc.
class SlurmJob:
    """
    Sets up and optionally submits slurm job scripts on HPC systems. Default settings for
    your HPC system can be added to dptools.hpc.hpc_defaults, or controlled using the
    CLI :doc:`../commands/set` command.

    .. role:: python(code)
        :language: python

    Args:
        sbatch_comment (str): Single line containing all Slurm #SBATCH parameters.
            e.g., :python:`'#SBATCH -J job -q regular -N 1 --time=11:00:00'`

        directories (list[str] or str): Path(s) to job submission directories.

        file_name (str): Name of submission script to write. Currently only supports .sh scripts,
            but support for others (e.g., .py) can be easily added upon request.

        zip_commands (bool): Set to True if each submission dir requires a unique set of commands.
            If True, then len(commands) must equal len(directories), and each item in commands
            is used for the respective directory index. i.e., job scripts are written as,

            .. code-block:: python

                for command, dir in zip(commands, directories):
                    write_script(command=command, submission_dir=dir)

            If False, then the same command(s) is used in all submission directories.

        **kwargs: Unpacked dict containing any env variables to set in submission script.
            e.g. :python:`kwargs = dict(TF_INTRA_OP_PARALLELISM_THREADS="1")` adds this
            line to .sh script,

            .. code-block:: bash

                export TF_INTRA_OP_PARALLELISM_THREADS=1
    """
    def __init__(self,
                 sbatch_comment,
                 commands="",
                 directories=".",
                 file_name="script.sh",
                 zip_commands=False,
                 **kwargs
                 ):

        self.sbatch = sbatch_comment
        self._zip = zip_commands
        self.commands = commands
        self.set_path_stuff(directories, file_name)
        self.set_text(**kwargs)

    def get_header(self):
        header = f"#!/usr/bin/env bash\n{self.sbatch}\n"
        return header

    def set_path_stuff(self, directories, file_name):
        if isinstance(directories, str):
            directories = [directories]
        self.directories = [os.path.abspath(d) for d in directories]
        self.file_name = file_name
        self.paths = [os.path.join(d, file_name) for d in self.directories]

    def set_text(self, **kwargs):
        header = self.get_header()

        exports = ""
        for k, v in kwargs.items():
            exports += f"export {k}={v}\n"

        body = "\n"
        if not self._zip and isinstance(self.commands, list):
            for comm in self.commands:
                body += comm + "\n"
        elif isinstance(self.commands, str):
            body += self.commands

        else:    # do not write commands to body if self._zip
            pass # in this case, self.commands contains
                 # list of unique commands for each submission dir

        self.text = header + exports + body

    @property
    def text(self):
        if self._zip:
            return self._text + next(self.comm_generator)
        return self._text

    @text.setter
    def text(self, new_text):
        self._text = new_text
        if self._zip:
            self.comm_generator = self.generate_commands()

    def generate_commands(self):
        for comm in self.commands:
            yield comm

    def write_script(self, path):
        with open(path, "w") as file:
            file.write(self.text)

    def write(self, sub=False):
        for directory, path in zip(self.directories, self.paths):
            self.write_script(path)
            if sub:
                os.chdir(directory)
                os.system(f"sbatch {self.file_name}")

    def submit(self):
        self.write(sub=True)



from dptools.simulate.simulations import Simulations, SPE, Opt, CellOpt, NVT, NPT
from dptools.simulate.calculator import DeepMD

__all__ = ["Simulations", "SPE", "Opt", "CellOpt", "NVT", "NPT", "DeepMD"]



from dptools.train.ensemble import SampleConfigs
from dptools.train.input import DeepInput, DeepInputs
from dptools.train.parity import EvaluateDP

__all__ = ["DeepInput", "DeepInputs", "EvaluateDP", "SampleConfigs"]



from dptools.cli import BaseCLI
from dptools.env import load, clear, clear_model, set_default_sbatch
from dptools.hpc import hpc_defaults
from dptools.simulate.parameters import reset_params


# TODO: Add reset params or reset {calculation_type} (e.g. nvt-md) in case
#       user screws up parameter_sets.yaml (also prevent this from happening)
class CLI(BaseCLI):
    """
    Reset models or sbatch parameters for environments made using ``dptools set ...``.

    :doc:`Complete documentation here<../commands/reset>`

    Examples:

    .. code-block:: console

        $ dptools reset all
        $ dptools reset -m water all
        $ dptools reset sbatch
    """

    help_info = "Reset model or sbatch params for default or labeled dptools env"
    def add_args(self):
        self.parser.add_argument(
            "thing",
            type=str,
            choices=("all", "sbatch", "model", "params"),
            help="Thing to reset (delete or set to default if exists)"\
                " (all, sbatch, model)"
        )
        self.parser.add_argument("-m", "--model-label", type=str, default=None,
                help="Label of specific model to use (see dptools set -h)")

    def main(self, args):
        if args.model_label:
            load(args.model_label)

        if args.thing == "all":
            clear(all)
        elif args.thing == "sbatch":
            set_default_sbatch(warn=False)
        elif args.thing == "model":
            clear_model()
        elif args.thing == "params":
            reset_params()



import os

from dptools.cli import BaseCLI
from dptools.train.input import DeepInputs


class CLI(BaseCLI):
    """
    Create deepmd-kit training set folder from ASE .db, .traj, or vasprun.xml files.

    :doc:`Complete documentaion here<../commands/input>`

    Examples:

    .. code-block:: console

        $ dptools input 00_system1.db 00_system2.db
        $ dptools input 0*_sys*.db # equivalent to above
        $ dptools input 00_system1/vasprun.xml 00_system2/vasprun.xml
        $ dptools input -p /path/to/dataset/folder 0*.db
    """
    help_info = "Set up deepmd-kit training input from ASE/VASP output"
    def add_args(self):
        self.parser.add_argument("inputs", nargs='+', metavar="input_file",
                help=".db, .traj, or vasprun.xml files")
        self.parser.add_argument("-n", type=int, default=None,
                help="Max number of images to take from each db")
        self.parser.add_argument("-p", "--path", type=str, default="./data",
                help="Specify path to dataset directory")
        self.parser.add_argument("-a", "--append", action="store_true",
                help="Append to dataset if system already exists in dataset directory")

    def main(self, args):
        self.names = []
        for inp in args.inputs:
            self.set_name(inp)
        path = os.path.abspath(args.path)
        thing = DeepInputs(args.inputs,
                           system_names=self.names,
                           path=path,
                           append=args.append,
                           n=args.n,
                           )

    def set_name(self, input_file):
        path, ext = os.path.splitext(input_file)
        if ext not in [".db", ".traj", ".xml"]:
            raise ValueError(f"Unrecognized input file type, {input_file}")
        elif ext == ".xml" and path.split("/")[-1] == "vasprun":
            name = path.split("/")[-2]
        else:
            name = path.split("/")[-1]

        if name in self.names:
            raise ValueError(f"Multiple inputs with same system name detected: {name}"\
                    ". Give each input file a unique name.")

        self.names.append(name)



"""
Module to generate parity plots to compare DP model predictions with corresponding
*ab-initio* values.
"""
import numpy as np
import matplotlib.pyplot as plt
import os

from dptools.utils import colors


class EvaluateDP:
    """
    Class to read deepmd test sets (created with CLI :doc:`../commands/input` command
    or :class:`~dptools.train.DeepInput`) and create parity plots for DP predictions.

    Args:
        test_sets (list[str] or str): Paths to deepmd test set folders. E.g.,
            ``'data/system1/test/set.000'``
            # TODO: Add support for other input types (.traj with vasp calculators, etc.)

        dp_graph (str): Path to deepmd model to use for DP predictions.

        per_atom (bool): If True, normalize all energies per number of atoms. False uses
            raw energies for parity plot and loss function evaluations.
    """

    def __init__(self, test_sets, dp_graph="graph.pb", per_atom=False):
        from deepmd.infer import DeepPot as DP
        self.dp = DP(dp_graph)
        if isinstance(test_sets, str):
            test_sets = [test_sets]

        self._per_atom = per_atom

        self.energies = []
        self.forces = []
        self.virials = []
        self.sets = test_sets
        for test_set in self.sets:
            energies, forces, virials = self.evaluate(test_set)
            self.energies.append(energies)
            self.forces.append(forces)
            if virials is not None:
                self.virials.append(virials)

        self.set_mse()

    def set_mse(self):
        self.all_mse = [[self.get_mse(e), self.get_mse(f)]
                         for e, f in zip(self.energies, self.forces)]
        if len(self.virials) > 0:
            self.all_v_mse = [self.get_mse(v) for v in self.virials]
            self.v_mse = self.get_mse(np.vstack(self.virials))

        self.mse = [self.get_mse(np.vstack(self.energies)),
                    self.get_mse(np.vstack(self.forces))]

    def evaluate(self, test_set):
        coord = np.load(f"{test_set}/coord.npy")
        cell = np.load(f"{test_set}/box.npy")
        atype = np.loadtxt(f"{test_set}/../type.raw", dtype=int)
        e_dp, f_dp, v_dp = self.dp.eval(coord, cell, atype)
        e_dp = e_dp.flatten()
        f_dp = f_dp.flatten()
        v_dp = v_dp.flatten()

        e_dft = np.load(f"{test_set}/energy.npy")
        f_dft = np.load(f"{test_set}/force.npy").flatten()
        if "virial.npy" in os.listdir(test_set):
            v_dft = np.load(f"{test_set}/virial.npy").flatten()
            virials = np.append(v_dft[:, np.newaxis], v_dp[:, np.newaxis], axis=1)
        else:
            virials = None

        energies = np.append(e_dft[:, np.newaxis], e_dp[:, np.newaxis], axis=1)
        if self._per_atom:
            n_atoms = len(atype)
            energies /= n_atoms
        forces = np.append(f_dft[:, np.newaxis], f_dp[:, np.newaxis], axis=1)
        return energies, forces, virials

    @staticmethod
    def get_mse(data):
        return np.mean((data[:, 0] - data[:, 1])**2)

    @staticmethod
    def get_rmse(data):
        return np.sqrt(np.mean((data[:, 0] - data[:, 1])**2))

    @staticmethod
    def get_mae(data):
        return np.mean(np.abs(data[:, 0] - data[:, 1]))

    @staticmethod
    def plot_yx(dft, ax):
        xrng = max(dft) - min(dft)
        xmin = min(dft) - 0.05 * xrng
        xmax = max(dft) + 0.05 * xrng
        ax.plot([xmin, xmax], [xmin, xmax], "--k", zorder=1)
        ax.set_xlim([xmin, xmax])
        ax.set_ylim([xmin, xmax])

    def plot_parity(self, data, label, color, loss="mse", ax=None, fancy=False, rasterized=False):
        if ax is None:
            ax = plt.gca()
        err = getattr(self, f"get_{loss.lower()}")(data)
        if not fancy:
            ax.plot(data[:, 0], data[:, 1], "o", ms=3, color=color, alpha=0.20,
                    rasterized=rasterized)
        else:
            density_scatter(data[:, 0], data[:, 1], ax=ax, zorder=10, rasterized=rasterized)
        self.plot_yx(data[:, 0], ax)
        ax.annotate(f"{loss.upper()} = {err:.3e}", xy=(0.1, 0.85),
                xycoords="axes fraction", fontsize=12)
        ax.set_ylabel(f"DP {label}", fontsize=14)
        ax.set_xlabel(f"DFT {label}", fontsize=14)
        ax.tick_params(labelsize=12)

    def plot(self, loss="mse", axs=None, xyz=False, fancy=False, save_file=None, rasterized=False):
        """
        Plot energy, force, and virial (if available) parity plots for DP predictions.

        Args:
            loss (str): {mse,mae,rmse} type of loss/error function to use for parity plot
                error annotation.
            axs (list[matplotlib.pyplot.Axes]): Optional, specific axes objects to plot on.
            xyz (bool): If True, plot all xyz force components in separate parity plots. False
                plots all components together (recommended, usually pointless to separate).
            fancy (bool): If True, create fancy density parity plot for force predictions.
        """
        n_plots = 2
        if axs is None:
            if len(self.virials) > 0:
                n_plots += 1
                v_data = np.vstack(self.virials)
            if xyz:
                n_plots += 2
        width = 0.5 + 4 * n_plots
        fig, axs = plt.subplots(1, n_plots, figsize=(width, 3.5))

        # TODO: Add ability to plot inidividual test sets
        #np.save("energies", np.array(self.energies))
        #np.save("forces", np.array(self.forces))
        e_data = np.vstack(self.energies)
        f_data = np.vstack(self.forces)

        e_units = "eV" if not self._per_atom else "eV/atom"
        self.plot_parity(e_data, f"Energy ({e_units})", colors[3], loss=loss, ax=axs[0])
        if not xyz:
            self.plot_parity(f_data, "Force (eV/Å)", colors[0], loss=loss, ax=axs[1], fancy=fancy,
                    rasterized=rasterized)
        else:
            self.plot_parity(f_data[::3, :], "F$_x$ (eV/Å)", colors[0],
                    loss=loss, ax=axs[1], fancy=fancy, rasterized=rasterized)

            self.plot_parity(f_data[1::3, :], "F$_y$ (eV/Å)", colors[0],
                    loss=loss, ax=axs[2], fancy=fancy, rasterized=rasterized)

            self.plot_parity(f_data[2::3, :], "F$_z$ (eV/Å)", colors[0],
                    loss=loss, ax=axs[3], fancy=fancy, rasterized=rasterized)

        if len(axs) in [3, 5]:
            self.plot_parity(v_data, "Virial", colors[2], loss=loss, ax=axs[-1],
                    rasterized=rasterized)
        plt.tight_layout()
        plt.subplots_adjust(wspace=0.3)
        if save_file:
            plt.savefig(save_file)
        else:
            plt.show()


def density_scatter(x, y, ax=None, bins=300, **kwargs):
    """
    Plot fancy density parity plot. Requires scipy package!

    Args:
        x (array-like): x-axis values to plot.
        y (array-like): y-axis values to plot.
        ax (matplotlib.axes.Axes): Axes object to plot on.
        bins (int): Number of bins to partition off x y values into. Passed to
            np.histogram2d(bins=bins).
        **kwargs: Any additional keyword-args for matplotlib.pyplot.scatter().
    """
    from scipy.interpolate import interpn
    if ax is None:
        fig, ax = plt.subplots()

    data, x_e, y_e = np.histogram2d(x, y, bins=bins, density=True)

    z = interpn((0.5 * (x_e[1:] + x_e[:-1]), 0.5 * (y_e[1:] + y_e[:-1])),
                data,
                np.vstack([x,y]).T,
                method="splinef2d",
                bounds_error=False)

    z[np.where(np.isnan(z))] = 0.0 # ignore div by 0 NaN

    idx = z.argsort()
    x, y, z = x[idx], y[idx], z[idx]

    ax.scatter(x, y, s=0.1, c=z, cmap="Spectral_r", **kwargs)



"""
Module for working with ensembles of DP models.
"""
from ase.io import read, write
import numpy as np
import random
import os

from dptools.utils import graph2typemap, next_color

class SampleConfigs:
    """
    Class for selecting new training configurations from snapshots of a molecular
    dynamnics trajectory (or some similar method). Uses an ensemble of models to
    calculate eps_t (max force prediction deviation), and then select new
    configurations within some specified tolerance such that new configurations
    belong to realistic but unexplored regions of configuration space.

    Follows the guidelines described by DP-GEN (full credit to the authors).
    For more details on eps_t and the methodologies used here, refer to:

        Y. Zhang, H. Wang, W. Chen, J. Zeng, L. Zhang, H. Wang and W. E,
        Comput. Phys. Commun., 2020, 253, 107206.

    Please cite the above reference if you use this script for training models in
    published work. Also, check out the DP-GEN GitHub for a robust, standalone package
    for automatically and efficiently training deepmd-kit MLPs.

    Args:
        configs (list[ase.Atoms] or str): List of atomic configurations from MD trajectory
            (or str to .traj, .xyz, etc. that contains configs) to sample from.
        graphs (list[str]): List of paths to ensemble of deepmd models (.pb files).
            e.g., ['00/graph.pb', '01/graph.pb', '02/graph.pb', '03/graph.pb']
        type_map (dict, optional): Dictionary mapping each atom type (symbol) to
            corresponding index. If None specified, infer from graph file.
        indices (str): Index slice to use for reading configs if str input supplied
            (used in command ase.io.read(config, index=indices)).
    """
    def __init__(self, configs, graphs, type_map=None, indices=":"):
        if isinstance(configs, str):
            self.configs = read(configs, index=indices)
        else:
            self.configs = configs
        if not type_map:
            type_map = graph2typemap(graphs[0])
        self.type_map = type_map
        self.graphs = graphs

    def get_dev(self):
        if "dev.npy" in os.listdir():
            old_dev = np.load("dev.npy")
            if len(old_dev) == len(self.configs): # only read dev if configs haven't changed
                print(f"Reading dev from {os.path.abspath('dev.npy')} ...")
                return old_dev

        from deepmd.infer import calc_model_devi
        from deepmd.infer import DeepPot as DP
        pos = np.array([a.get_positions().flatten() for a in self.configs])
        cell = np.array([a.cell.array.flatten() for a in self.configs])
        types = [self.type_map[a.symbol] for a in self.configs[0]]

        models = [DP(g) for g in self.graphs]

        dev = calc_model_devi(pos, cell, types, models, nopbc=False)[:, 4]
        np.save("dev.npy", dev)
        return dev

    def sample(self, lo=0.05, hi=0.35, n=300):
        """
        Select n new training configurations with lo < eps_t < hi.

        .. note::

            ``hi`` must be chosen carefully to allow for the selection of
            configurations that belong to underexplored regions of
            configuration space, but not set so high that nonsensical,
            unphysical configs are chosen.

        Args:
            lo (float): Lower bound eps_t tolerance for sampling.
            hi (float): Upper bound eps_t tolerance for sampling.
            n (int): Maximum number of configurations to select. If the number of
                configs within lo and hi is < n, then all configs between lo and hi
                are returned.

        Returns:
            new_configs (list[ase.Atoms]): Configurations with eps_t within the
                specified tolerance criteria.
        """
        if hi > 0.5:
            warn = """
               ----------------------------------------------------------------------------------
              | WARNING: High upper bound eps_t tolerance specified!                             |
              |----------------------------------------------------------------------------------|
              | Large eps_t might correspond with unphysical (bad) configurations.               |
              | IF YOU USE THESE CONFIGS FOR TRAINING, MAKE SURE YOUR DFT CALCULATION CONVERGES! |
              | You will regret this if you do not heed this warning, trust me.                  |
              |__________________________________________________________________________________|
               """ # large obnoxious warning box for emphasis
            print(warn)

        self.dev = self.get_dev()
        i_configs = np.where(np.logical_and(self.dev>=lo, self.dev<=hi))[0]
        n_sample = n if n < len(i_configs) else len(i_configs)
        i_new_configs = random.sample(list(i_configs), n_sample)
        new_configs = [self.configs[i] for i in i_new_configs]
        return new_configs

    def plot(self, dev=None, steps=False, ax=None, color=None, label=None):
        """
        Create kernel density estimation plot (fancy smooth histogram) for
        all eps_t values in configs input.

        Requires seaborn python package to be installed!
        https://seaborn.pydata.org

        Args:
            dev (array-like): Optional list of dev values to plot, calculate if None given.
            steps (bool): If False, plot histogram of dev values. If True, plot dev versus
                step number (divided by write_freq).
            ax (Axes): Specific mpl Axes to plot on.
            color (str): Color to use for plot.
            label (str): Label to use for legend entry.
        """
        import matplotlib.pyplot as plt
        import seaborn as sns
        if dev is None:
            if hasattr(self, "dev"):
                dev = self.dev
            else:
                dev = self.get_dev()
        ax = plt.gca() if ax is None else ax
        color = next_color() if color is None else color

        if steps:
            ax.set_xlabel("Steps / write_freq", fontsize=14)
            ax.set_ylabel("$\epsilon_t$ (eV/Å)", fontsize=14)
            plt.plot(np.arange(len(dev)), dev, '-', color=color, label=label or "")
        else:
            sns.kdeplot(dev, fill=True, color=color, label=label or "")
            ax.set_ylabel("Density", fontsize=14)
            ax.set_xlabel("$\epsilon_t$ (eV/Å)", fontsize=14)
        ax.locator_params('both', nbins=5)
        ax.tick_params(labelsize=12)
        plt.tight_layout()
        return ax



import os
from ruamel.yaml import YAML
from ase.io import read
import numpy as np

from dptools.simulate import Simulations
from dptools.simulate.parameters import get_parameter_sets
from dptools.utils import read_type_map
from dptools.cli import BaseCLI
from dptools.env import get_dpfaults, load


class CLI(BaseCLI):
    """
    Run LAMMPS simulations using trained DP models.

    :doc:`Complete documentation here<../commands/run>`

    Examples:

    .. code-block:: console

        $ dptools run opt start.traj # simple atomic position optimization
        $ dptools run cellopt start.traj # simple unit cell optimization
        $ dptools run /path/to/params.yaml start.traj # custom param file simulation
        $ dptools run -s eos 0*/start.traj # submit slurm job eos simulations on multiple structures
        $ dptools run -s -m water nvt-md start.traj # submit slurm nvt-md run using set water model
    """
    help_info = "Run simulation using trained DP model "\
            "(USE COMMAND 'dptools set path/to/graph.pb' first)"

    def add_args(self):
        self.parser.add_argument(
            "calculation",
            type=str,
            help="Type of calculation to run "\
                    "(spe, opt, cellopt, nvt-md, npt-md, eos, vib, or params.yaml)",
        )
        self.parser.add_argument(
            "structure",
            nargs="+",
            help="File containing structure to run calculation on (.traj, .xyz, .cif, etc.)"
        )
        self.parser.add_argument("-m", "--model-label", type=str, default=None,
                help="Label of specific model to use (see dptools set -h)")
        self.parser.add_argument("-s", "--submit", action="store_true",
                help="Automatically submit job(s) to train model(s) once input has been created")
        self.parser.add_argument("-g", "--generate-input", action="store_true",
                help="Only setup calculation and generate input files but do not run calculation")
        #self.parser.add_argument("-p", "--path", nargs=1, type=str, default="./",
        #        help="Specify path to write simulation files and results to")
        self.parser.add_argument("-o", "--output", type=str, default="{calculation}.traj",
                help="Name of file to write calculation output to")

    def main(self, args):
        self.set_model(args.model_label)
        self.set_params(args.calculation)
        self.set_structures(args.structure)
        if args.output == "{calculation}.traj": # replace default placeholder name
            args.output = f"{self.calc_type}.traj"
        self.file_out = args.output

        if args.submit:
            self.submit_jobs(sub=True)
        elif args.generate_input:
            self.submit_jobs(sub=False)
        else:
            self.run()

    def set_params(self, calc_arg):
        """
        Set simulation parameters either from simulation keyword or params.yaml file.

        Note:
            If using parameter file, you must at least retain the .yaml extension.

        Args:
            calc_arg (str): simulation keyword (e.g. `opt`) or path to params.yaml file.
        """

        if calc_arg.endswith(".yaml"):
            calc_arg = os.path.abspath(calc_arg)
            with open(calc_arg) as file:
                params = YAML().load(file.read())
        else:
            param_sets = get_parameter_sets()
            params = param_sets[calc_arg]
        self.calc_type = params.pop("type").split(".", 1)[0]
        self.params = params
        self.calc_arg = calc_arg # needed for rewriting dptools command in job submission script

    def set_model(self, model_label):
        """
        Load specific model (if model_label is not None) and set corresponding graph and type_map.

        Args:
            model_label (str or None): Label of specific model environment to load.
        """
        # NOTE: NEED TO SET THE LABEL BEFORE CALLING get_dpfaults
        #  Also, I suppose it's not really get_dpfaults if it can load custom envs
        if model_label:
            if os.path.isfile(model_label) and model_label.endswith(".pb"):
                err = "\nCan not load .pb files for 'dptools run -m ...' (long story)"\
                      ". Give model a label and use with set command! Example:\n\n\t"\
                      f"dptools set -m label {model_label}\n"
                raise ValueError(err)

            load(model_label)

        self.graph, self.type_map = get_dpfaults()
        self._label = model_label # need for submit_jobs()

    def set_structures(self, structures):
        """
        Read and set structure inputs as ase.Atoms objects and set corresponding dirs.
        Only reads last index unless calc_type == spe, in which case single points are ran
        on all images of structure file.

        Args:
            structures (list of str): Paths to structure inputs to run simulations
                on (.traj, .xyz, .cif, etc.).
        """

        index = -1
        self.structures = [os.path.abspath(s) for s in structures]
        if len(structures) == 1:
            dirs = ["."]
            if self.calc_type == "spe":
                index = ":"
        else:
            dirs = [os.path.dirname(s) for s in self.structures]
            if len(np.unique(dirs)) != len(self.structures):
                # FIXME: Results are overwritten if multiple structure inputs are in the same dir
                raise Exception("Can't resolve inputs, harass me to fix this")

        self.atoms = [read(s, index=index) for s in self.structures]
        self.dirs = dirs

    def submit_jobs(self, sub=True):
        """
        Setup and (optionally) submit slurm jobs on all structure inputs.

        Args:
            sub (bool, optional): Submits slurm jobs if True, else only writes
                input files without submitting.
        """
        from dptools.hpc import SlurmJob
        hpc_info = get_dpfaults(key="sbatch")
        sbatch_comment = hpc_info.pop("SBATCH_COMMENT")

        f_arg = f"-o {self.file_out}"
        m_arg = f"-m {self._label} " if self._label else ""
        comm_base = f"dptools run {m_arg}{f_arg} {self.calc_arg} "
        commands = [comm_base + os.path.basename(s) for s in self.structures]
        jobs = SlurmJob(sbatch_comment,
                        commands=commands,
                        directories=self.dirs,
                        file_name="dptools.run.sh",
                        zip_commands=True,
                        **hpc_info
                        )
        jobs.write(sub=sub)

    def run(self):
        """
        Sequentially setup and run simulations on all structure inputs.
        """
        wd = os.getcwd()
        for atoms, d in zip(self.atoms, self.dirs):
            os.chdir(d)
            sim = Simulations[self.calc_type](
                    atoms,
                    self.graph,
                    type_map=read_type_map(self.type_map),
                    file_out=self.file_out,
                    path=".",
                    **self.params
                    )

            sim.run()
            os.chdir(wd)



from dptools import __version__
from setuptools import setup
from pathlib import Path

version = __version__

HERE = Path(__file__).parent

README = (HERE / "README.md").read_text()

install_requires = (HERE / 'requirements.txt').read_text().splitlines()

python_requires = '>=3.6' # lots of f-string usage

packages = ['dptools', 'dptools.simulate', 'dptools.cli', 'dptools.train']
package_data = {'dptools': ['simulate/parameter_sets.yaml', 'train/in.json']}

setup(name='dpmdtools', # dptools taken on PyPI :(
      version=version,
      description='DPTools: CLI toolkit and python library for working with deepmd-kit.',
      long_description=README,
      long_description_content_type='text/markdown',
      url='https://github.com/tysours/DPTools',
      maintainer='Ty Sours',
      maintainer_email='tsours@ucdavis.edu',
      license='MIT',
      packages=packages,
      python_requires=python_requires,
      install_requires=install_requires,
      #extras_require=extras_require,
      package_data=package_data,
      entry_points={'console_scripts': ['dptools=dptools.cli:main']},
      classifiers=[
          'Development Status :: 3 - Alpha',
          'License :: OSI Approved :: MIT License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3',
          'Intended Audience :: Science/Research',
          'Topic :: Scientific/Engineering :: Chemistry'],
      project_urls={
          'Documentation': 'http://dptools.readthedocs.io/',
          'Source': 'https://github.com/tysours/DPTools',
          'Tracker': 'https://github.com/tysours/DPTools/issues',
          },
      )



import re
import socket

sim_queue_file = 'sims'

def _gen_simulation(sim_queue_file):
    with open(sim_queue_file) as file:
        simulations = [l for line in file for l in line.split()]
    return simulations

s = _gen_simulation(sim_queue_file)
hostname = 'cori09'
host = re.sub("[^a-z]*", "", socket.gethostname())
host = re.sub("[^a-z]*", "", hostname)




"""
Functions for controlling different dptools env configurations
(deepmd models, sbatch parameters, etc.).
i.e. controls things set with CLI :doc:`../commands/set` command.

Mostly indented to be used behind the scenes when using CLI commands, but theoretically could
be used in your python scripts as well. Slightly confusing usage with global env_file (sorry),
so use at your own risk.
"""
import os
import re
import json
import socket
import dotenv

from dptools.utils import typemap2str, graph2typemap
from dptools.hpc import hpc_defaults

basedir = os.path.abspath(os.path.dirname(__file__))
default_env_file = os.path.join(basedir, ".env")
env_file = default_env_file


def set_env(key, value):
    """Set key-value to global env file."""
    dotenv.set_key(env_file, key, value)


def get_env():
    """Get all key-values from global env file."""
    values = dotenv.dotenv_values(env_file)
    return values


def load(label):
    """Sets global env file if different from default .env. Used to load custom
    envs essentially.
    """
    if label:
        global env_file
        env_file = default_env_file + "." + label


def get_dpfaults(key="model"):
    """
    Like defaults but for dp (haha... ha..). Loads specific key-values from env file.

    Args:
        key (str): ``(model, ensemble, sbatch)`` Gets env information from env file.

            * if key is set to model, return env's deepmd model path and type map
            * if key is set to ensemble, return all model paths belonging to env's ensemble
            * if key is set to sbatch, return env's Slurm settings
    """
    default_vals = get_env()

    if key == "model":
        keys = ["DPTOOLS_MODEL", "DPTOOLS_TYPE_MAP"]
        defaults = tuple([default_vals.get(k) for k in keys])

    elif key == "ensemble":
        keys = ["DPTOOLS_TYPE_MAP",
                "DPTOOLS_MODEL",
                "DPTOOLS_MODEL2",
                "DPTOOLS_MODEL3",
                "DPTOOLS_MODEL4"]
        defaults = tuple([default_vals.get(k) for k in keys])

    elif key == "sbatch":
        keys = ["SBATCH_COMMENT",
                "OMP_NUM_THREADS",
                "TF_INTRA_OP_PARALLELISM_THREADS",
                "TF_INTER_OP_PARALLELISM_THREADS"]

        if not default_vals.get(keys[0], None):
            set_default_sbatch()
            default_vals = get_env()
        defaults = {k: default_vals[k] for k in keys}
    return defaults


def set_default_sbatch(warn=True):
    """
    Sets Slurm parameters from dptools.hpc.hpc_defaults. Mostly used by CLI when
    no Slurm info has been set to env.
    """
    host = re.sub("[^a-z]*", "", socket.gethostname())
    try:
        for k, v in hpc_defaults[host].items():
            set_env(k, str(v))
        if warn:
            print("WARNING: setting default HPC parameters to env")
            print("-" * 64)
            print("\nSettings:")
            for k, v in hpc_defaults[host].items():
                print(k, "=", v)
            print("-" * 64)
    except KeyError as exc:
        raise Exception("Host unrecognized and no default HPC parameters found."\
            "\nUse 'dptools set script.sh' with desired #SBATCH comment in script.sh") from exc


def clear(keys):
    """
    Clear specific key-values (or entire env) for loaded global env file.
    """
    if keys is all:
        os.remove(env_file)
    else:
        vals = get_env()
        for key in keys:
            if vals.get(key):
                dotenv.unset_key(env_file, key)


def clear_model():
    """
    Clear model related key-values from loaded global env file.
    """
    keys = [
            "DPTOOLS_TYPE_MAP",
            "DPTOOLS_MODEL",
            "DPTOOLS_MODEL2",
            "DPTOOLS_MODEL3",
            "DPTOOLS_MODEL4",
        ]
    clear(keys)


def set_model(model, n_model=""):
    """
    Set path to deepmd model .pb file to use during simulations evoked by CLI
    :doc:`../commands/run` command.
    """
    if not n_model:
        clear_model()
    graph = os.path.abspath(model)
    set_env(f"DPTOOLS_MODEL{n_model}", graph)
    if not n_model: # only write type_map once if setting ensemble of models
        type_map = graph2typemap(graph)
        type_map_str = typemap2str(type_map)
        set_env("DPTOOLS_TYPE_MAP", type_map_str)


def set_sbatch(script):
    """
    Reads script and searches for any #SBATCH line to set to env file.

    Args:
        script (str): Path to .sh script with #SBATCH comments to set Slurm params.
    """
    with open(script) as file:
        lines = [l.strip() for l in file]
    sbatch_vars = []
    for l in lines:
        if l.startswith("#SBATCH"):
            sbatch_vars.extend(l.split()[1:])
        elif l.startswith("export"):
            var = l.split()[1]
            set_env(*var.split("="))
    sbatch_comment = "#SBATCH " + " ".join(sbatch_vars)
    set_env("SBATCH_COMMENT", sbatch_comment)


def set_params(params):
    """
    Save set of simulation parameters to run with CLI :doc:`../commands/run` command.
    """
    from dptools.simulate.parameters import set_parameter_set
    set_parameter_set(params)


def set_training_params(in_json):
    """
    Save deepmd-kit training parameters to use with CLI :doc:`../commands/train` command.
    """
    if isinstance(in_json, str):
        with open(in_json) as file:
            in_json = json.loads(file.read())
    if not in_json.get("model"):
        # check if correct param file/dict was given before overwriting default file
        raise KeyError("No model parameters found in json file.")

    # no need to keep these values since they will be overwritten anyway
    in_json["model"]["type_map"] = []
    in_json["training"]["training_data"]["systems"] = []
    in_json["training"]["validation_data"]["systems"] = []

    path = os.path.join(basedir, "train/in.json")
    with open(path, "w") as file:
        file.write(json.dumps(in_json, indent=4))



import argparse
import re
from importlib import import_module
from textwrap import dedent, fill

from dptools import __version__

commands = (
    "input",
    "train",
    "parity",
    "run",
    "sample",
    "convert",
    "set",
    "get",
    "reset",
    "info",
)

class BaseCLI:
    """
    Base class for CLI commands. More or less just a template for reference.

    Args:
        parser (argparse.ArgumentParser): argparse parser for parsing CLI command arguments
    """
    help_info = ""

    def __init__(self, parser):
        self.parser = parser

    def add_args(self):
        '''Command specific arguments'''
        return

    def main(self, args):
        '''Command specific main method'''
        return


class MyFormatter(argparse.ArgumentDefaultsHelpFormatter):
    """
    HelpFormatter that replaces reST refs in docstrings with readthedocs URLs
    to print to console. Also dedents and retains docstring format.
    """

    def _fill_text(self, text, width, indent):
        # replaces e.g. :doc:`text<ref_path>`
        pattern = r":[a-z]+:`[\w ]+<[.\w/]+>`"
        subpattern_link = r"<../[a-z/]+>"
        subpattern_text = r"`[\w ]+<"

        url = ": https://dptools.rtfd.io/en/latest/"

        matches = re.findall(pattern, text)
        for match in matches:
            keep = re.search(subpattern_text, match).group()[1:-1]
            sub = re.search(subpattern_link, match).group()[4:-1]
            new = f"{keep}{url}{sub}.html"
            text = text.replace(match, new)

        # replaces e.g. .. command:: text
        pattern = r"[ ]*[.]{2} [\w-]+::[\w ]*\n\n"
        while re.search(pattern, text):
            text = re.sub(pattern, "", text)

        return dedent(text)


def main():
    parser = argparse.ArgumentParser(prog="dptools",
                                     description="DPTools CLI for doing stuff with deepmd-kit\n\n"\
                                             "Complete documentation available at: "\
                                             "https://dptools.rtfd.io",
                                     formatter_class=MyFormatter,
                                     )
    parser.add_argument("--version", action="version", version=__version__)
    subparsers = parser.add_subparsers(dest="command")
    # TODO: Add logging
    command_clis = {}
    for comm in commands:
        mod = "dptools.cli." + comm
        CLI = import_module(mod).CLI
        doc = CLI.__doc__
        subparser = subparsers.add_parser(comm, help=CLI.help_info, description=doc,
                formatter_class=MyFormatter)
        cli = CLI(subparser)
        cli.add_args()
        command_clis[comm] = cli

    parsed_args = parser.parse_args()
    command_clis[parsed_args.command].main(parsed_args)
